"""Formatting models."""

from __future__ import annotations

from enum import Enum

from pydantic import BaseModel, Field


class ClickColor(Enum):
    """Colors supported by click 8.3."""

    black = "black"
    red = "red"
    green = "green"
    yellow = "yellow"
    blue = "blue"
    magenta = "magenta"
    cyan = "cyan"
    white = "white"
    bright_black = "bright_black"
    bright_red = "bright_red"
    bright_green = "bright_green"
    bright_yellow = "bright_yellow"
    bright_blue = "bright_blue"
    bright_magenta = "bright_magenta"
    bright_cyan = "bright_cyan"
    bright_white = "bright_white"
    reset = "reset"


class FormatObject(BaseModel):
    """Objects for output formatting."""

    available_updates: str
    last_update: str
    matched_criteria: str
    matched_criteria_short: str
    score: int
    status_text: str
    update_count: int
    # Excluded fields won't be usable in custom format
    state_color: str = Field(exclude=True)
    state_name: str = Field(exclude=True)



"""Config module."""

from pathlib import Path
from tomllib import TOMLDecodeError
from tomllib import load as toml_load
from typing import Any, Self

from pydantic import BaseModel, ConfigDict, Field, ValidationError, computed_field, field_validator, model_validator

from siun.errors import ConfigError
from siun.models import (
    CRITERION_REGISTRY,
    ClickColor,
    CriterionAvailable,
    CriterionCount,
    CriterionCustom,
    CriterionPattern,
    NewsProvider,
    V2Criterion,
    V2Threshold,
)
from siun.notification import UpdateNotification
from siun.providers import UPDATE_PROVIDER_REGISTRY, UpdateProvider, UpdateProviderPacman
from siun.util import get_default_config_dir, get_default_state_dir, is_path_world_writable


def get_default_thresholds() -> list[V2Threshold]:
    """Backwards compatible default thresholds."""
    return [
        V2Threshold(name="critical", score=3, color=ClickColor.red, text="Updates required"),
        V2Threshold(name="warning", score=2, color=ClickColor.yellow, text="Updates recommended"),
        V2Threshold(name="available", score=1, color=ClickColor.green, text="Updates available"),
    ]


def get_default_criteria() -> list[V2Criterion]:
    """Backwards compatible default criteria."""
    return [
        CriterionAvailable(name="available", weight=1),
        CriterionPattern(name="pattern", weight=1, pattern="^archlinux-keyring$|^linux$|^pacman.*$"),
        CriterionCount(name="count", weight=1, count=15),
    ]


def get_default_update_providers() -> list[UpdateProvider]:
    """Get default update providers."""
    return [UpdateProviderPacman()]


class SiunConfig(BaseModel):
    """Config struct."""

    update_providers: list[UpdateProvider] = Field(default_factory=get_default_update_providers)
    cache_min_age_minutes: int = Field(default=30)
    v2_thresholds: list[V2Threshold] = Field(default_factory=get_default_thresholds)
    v2_criteria: list[V2Criterion] = Field(default_factory=get_default_criteria)
    custom_format: str = Field(default="$status_text: $available_updates")
    state_dir: Path = Field(default_factory=get_default_state_dir)
    notification: UpdateNotification | None = Field(default=None)
    news: list[NewsProvider] = Field(default=[])

    @computed_field
    @property
    def sorted_thresholds(self) -> list[V2Threshold]:
        """
        Sort thresholds by descending score.

        This makes it easier to find the highest threshold that matches later.
        """
        return sorted(self.v2_thresholds, key=lambda item: item.score, reverse=True)

    @computed_field
    @property
    def mapped_thresholds(self) -> dict[str, V2Threshold]:
        """Map thresholds to threshold name."""
        return {t.name: t for t in self.v2_thresholds}

    @model_validator(mode="before")
    @classmethod
    def handle_deprecated_fields(cls, data: Any) -> Any:
        """Check for deprecated fields."""
        if not isinstance(data, dict):
            return data

        has_thresholds = "thresholds" in data
        has_criteria = "criteria" in data
        has_cmd_available = "cmd_available" in data
        has_state_file = "state_file" in data
        if not has_thresholds and not has_criteria and not has_cmd_available:
            return data

        message_parts = ["Found deprecated config fields: "]
        if has_thresholds:
            message_parts.append("- 'thresholds' have been deprecated in favour of 'v2_thresholds'")
        if has_criteria:
            message_parts.append("- 'criteria' have been deprecated in favour of 'v2_criteria'")
        if has_cmd_available:
            message_parts.append("- 'cmd_available' has been deprecated in favour of 'update_providers'")
        if has_state_file:
            message_parts.append("- 'state_file' has been deprecated in favour of 'state_dir'")

        raise ValueError("\n".join(message_parts))

    @model_validator(mode="after")
    def notification_must_reference_threshold(self) -> Self:
        """Check if notification threshold exists."""
        value = getattr(self, "notification", None)
        thresholds: list[str] = [t.name for t in getattr(self, "v2_thresholds", [])]

        if value and value.threshold not in thresholds:
            message = f"notification.threshold must be one of: {', '.join(thresholds)}"
            raise ValueError(message)

        return self

    @field_validator("v2_criteria")
    def transform_criteria(cls, value: list[V2Criterion]) -> list[V2Criterion]:
        """Transform criteria to subclasses of V2Criterion."""
        registry = CRITERION_REGISTRY
        custom_cls = CriterionCustom
        return [(registry.get(crit.name) or custom_cls)(**crit.model_dump(exclude={"name_short"})) for crit in value]

    @field_validator("v2_thresholds")
    def thresholds_must_have_unique_name(cls, value: list[V2Threshold]) -> list[V2Threshold]:
        """
        Check if all thresholds have a unique name.

        The name doubles as ID.
        """
        unique_names = {obj.name for obj in value}
        if len(unique_names) != len(value):
            message = "each threshold must have a unique name."
            raise ValueError(message)

        return value

    @field_validator("update_providers")
    def transform_update_providers(cls, value: list[UpdateProvider]) -> list[UpdateProvider]:
        """Transform update provider to subclasses of UpdateProvider."""
        registry = UPDATE_PROVIDER_REGISTRY
        transformed_value: list[UpdateProvider] = []
        for provider in value:
            try:
                transformed_value.append(registry[provider.name](**provider.model_dump()))
            except KeyError as error:
                message = f"unknown update provider: {provider.name}\nAvailable providers: {', '.join(registry.keys())}"
                raise ValueError(message) from error

        return transformed_value

    model_config = ConfigDict(extra="allow")


def _read_config(config_path: Path) -> dict[str, Any]:  # pragma: no cover
    """Read config from disk."""
    with Path.open(config_path, "rb") as file_obj:
        return toml_load(file_obj)


def _format_error_loc(err_loc: tuple[int | str, ...]):
    if err_loc:
        return f"'{'.'.join(str(loc) for loc in err_loc)}': "

    return ""


def _format_pydantic_error(error: ValidationError):
    return "; ".join(f"{_format_error_loc(err['loc'])}{err['msg']}" for err in error.errors())


def get_config(config_path: Path | None = None) -> SiunConfig:
    """Get config from defaults and user supplied values."""
    if config_path is None:
        config_path = get_default_config_dir() / Path("config.toml")

    if is_path_world_writable(config_path):
        message = (
            f"Config file '{config_path}' is world-writable. Please change its permissions to be more restrictive."
        )
        raise ConfigError(message, config_path)

    config_dict: dict[str, Any] = {}
    if config_path.exists() and config_path.is_file():
        try:
            config_dict = _read_config(config_path)
        except OSError as error:
            message = f"failed to open config file for reading: {error}"
            raise ConfigError(message, config_path) from error

        except TOMLDecodeError as error:
            message = f"provided config file not valid: {error}"
            raise ConfigError(message, config_path) from error

    else:
        message = f"config file not found: {config_path}\nPlease create a configuration file."
        raise ConfigError(message, config_path)

    try:
        config = SiunConfig(**config_dict)
    except ValidationError as error:
        message = f"invalid configuration: {_format_pydantic_error(error)}"
        raise ConfigError(message, config_path) from error

    return config



"""Collection of test fixtures."""

import datetime
from pathlib import Path
from unittest import mock

import pytest

from siun.config import SiunConfig, get_default_thresholds
from siun.models import (
    CriterionAvailable,
    CriterionCount,
    CriterionPattern,
    FormatObject,
    PackageUpdate,
    Updates,
    V2Threshold,
)
from siun.providers import UpdateProvider, UpdateProviderPacman


class DummyProvider(UpdateProvider):
    """Dummy provider for testing."""

    name: str = "dummy"

    def fetch_updates(self):
        """Return dummy updates."""
        return []


@pytest.fixture(scope="module")
def default_update_providers():
    """Provide default update provider."""
    return [UpdateProviderPacman()]


@pytest.fixture(scope="module")
def default_thresholds():
    """Provide default thresholds."""
    return get_default_thresholds()


@pytest.fixture(scope="module")
def default_config(default_thresholds):
    """Provide default config."""
    return SiunConfig(
        update_providers=[UpdateProvider(name="pacman")],
        cache_min_age_minutes=30,
        v2_thresholds=default_thresholds,
        v2_criteria=[
            CriterionAvailable(name="available", weight=1),
            CriterionPattern(name="pattern", weight=1, pattern="^archlinux-keyring$|^linux$|^pacman.*$"),
            CriterionCount(name="count", weight=1, count=15),
        ],
        custom_format="$status_text: $available_updates",
        state_dir=Path("/tmp/siun-tests/.local/state/siun"),  # noqa: S108
        news=[],
    )


@pytest.fixture(scope="module")
def config_w_notification(default_thresholds):
    """Provide default config."""
    return SiunConfig(
        update_providers=[UpdateProvider(name="pacman")],
        cache_min_age_minutes=30,
        v2_thresholds=default_thresholds,
        v2_criteria=[
            CriterionAvailable(name="available", weight=1),
            CriterionPattern(name="pattern", weight=1, pattern="^archlinux-keyring$|^linux$|^pacman.*$"),
            CriterionCount(name="count", weight=1, count=2),
        ],
        custom_format="$status_text: $available_updates",
        state_dir=Path("/tmp/siun-tests/.local/state/siun"),  # noqa: S108
        notification={"title": "siun test notification", "threshold": "available"},
    )


@pytest.fixture(scope="module")
def config_w_notification_threshold(default_thresholds):
    """Provide default config."""
    return SiunConfig(
        update_providers=[UpdateProvider(name="pacman")],
        cache_min_age_minutes=30,
        v2_thresholds=default_thresholds,
        v2_criteria=[
            CriterionAvailable(name="available", weight=1),
            CriterionPattern(name="pattern", weight=1, pattern="^archlinux-keyring$|^linux$|^pacman.*$"),
            CriterionCount(name="count", weight=1, count=2),
        ],
        custom_format="$status_text: $available_updates",
        state_dir=Path("/tmp/siun-tests/.local/state/siun"),  # noqa: S108
        notification={"title": "siun test notification", "threshold": "warning"},
    )


@pytest.fixture(scope="module")
def v2_config_w_custom_format(default_thresholds):
    """Provide default config."""
    return SiunConfig(
        update_providers=[UpdateProvider(name="pacman")],
        cache_min_age_minutes=30,
        v2_thresholds=[V2Threshold(score=1, name="available", text="Updates available")],
        v2_criteria=[
            CriterionAvailable(name="available", weight=1),
            CriterionCount(name="count", weight=1, count=15),
            CriterionPattern(name="pattern", weight=1, pattern="^archlinux-keyring$|^linux$|^pacman.*$"),
        ],
        custom_format="$status_text: $available_updates",
        state_dir=Path("/tmp/siun-tests/.local/state/siun"),  # noqa: S108
    )


@pytest.fixture
def state_stale(default_thresholds):
    """Expired Update state."""
    return Updates(
        criteria_settings=[
            CriterionAvailable(name="available", weight=1),
            CriterionCount(name="count", weight=1, count=2),
        ],
        thresholds_settings=default_thresholds,  # ty: ignore[unknown-argument]
        available_updates=[],
        last_update=datetime.datetime.now(tz=datetime.UTC) - datetime.timedelta(days=1),
    )


@pytest.fixture
def notification_mock():
    """Provide a mock notification object for tests."""

    def _make(threshold="warning", urgency=None):
        notification = mock.Mock()
        notification.threshold = threshold
        notification.urgency = urgency
        notification.hints = {}
        notification.fill_templates = mock.Mock()
        notification.show = mock.Mock()
        return notification

    return _make


@pytest.fixture(scope="module")
def updates_single():
    return PackageUpdate(name="siun", provider="name")


@pytest.fixture
def updates_multiple(count: int = 2):
    updates = []
    for i in range(count):
        updates.append(PackageUpdate(name=f"package_{i}", provider="name"))
    return updates


@pytest.fixture(scope="module")
def format_object_ok():
    """FormatObject with 'Ok' status."""
    return FormatObject(
        available_updates="",
        last_update="2025-01-01T00:00:00+00:00",
        matched_criteria="",
        matched_criteria_short="",
        score=0,
        status_text="Ok",
        update_count=0,
        state_color="green",
        state_name="Ok",
    )


@pytest.fixture(scope="module")
def format_object_available():
    """FormatObject with 'Updates available' status."""
    return FormatObject(
        available_updates="",
        last_update="2025-01-01T00:00:00+00:00",
        matched_criteria="",
        matched_criteria_short="",
        score=0,
        status_text="Updates available",
        update_count=0,
        state_color="blue",
        state_name="Updates available",
    )


@pytest.fixture(scope="module")
def format_object_recommended():
    """FormatObject with 'Updates recommended' status."""
    return FormatObject(
        available_updates="",
        last_update="2025-01-01T00:00:00+00:00",
        matched_criteria="",
        matched_criteria_short="",
        score=0,
        status_text="Updates recommended",
        update_count=0,
        state_color="yellow",
        state_name="Updates recommended",
    )


@pytest.fixture(scope="module")
def format_object_required():
    """FormatObject with 'Updates required' status."""
    return FormatObject(
        available_updates="",
        last_update="2025-01-01T00:00:00+00:00",
        matched_criteria="",
        matched_criteria_short="",
        score=0,
        status_text="Updates required",
        update_count=0,
        state_color="red",
        state_name="Updates required",
    )


@pytest.fixture
def format_object_factory():
    """Build factory for FormatObject instances with custom attributes."""

    def _make(**kwargs):
        defaults = {
            "available_updates": "",
            "last_update": "2025-01-01T00:00:00+00:00",
            "matched_criteria": "",
            "matched_criteria_short": "",
            "score": 0,
            "status_text": "Ok",
            "update_count": 0,
            "state_color": "green",
            "state_name": "Ok",
        }
        defaults.update(kwargs)
        return FormatObject(**defaults)

    return _make


@pytest.fixture(autouse=True, scope="class")
def patch_is_path_world_writable(request):
    """Mock is_path_world_writable to always return False to avoid early exit."""
    with mock.patch("siun.config.is_path_world_writable", return_value=False):
        yield



"""Threshold models."""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict

from siun.models.formatting import ClickColor


class V2Threshold(BaseModel):
    """Version 2 of the Threshold struct."""

    score: int
    name: str
    text: str
    color: ClickColor = ClickColor.reset

    model_config = ConfigDict(extra="forbid")



"""CLI utils."""

from collections.abc import Callable
from pathlib import Path
from typing import Any

import click

from siun.config import SiunConfig, get_config
from siun.errors import (
    ConfigError,
    SiunCLIError,
)


def common_options(f: Callable[..., Any]) -> Callable[..., Any]:
    """Define common options for CLI commands."""
    f = click.option(
        "--config-path",
        "-C",
        type=click.Path(exists=True, dir_okay=False, readable=True, path_type=Path),
        help="Override config file location.",
    )(f)
    return f


def load_config_or_exit(config_path: Path | None) -> SiunConfig:
    """Load configuration or exit with error."""
    try:
        config = get_config(config_path)
    except ConfigError as error:
        message = f"{error.message}; config path: {error.config_path}"
        raise SiunCLIError(message) from error
    return config



"""Generic package update provider."""

from __future__ import annotations

import subprocess

from pydantic import ConfigDict

from siun.errors import UpdateProviderError
from siun.models import PackageUpdate
from siun.providers.base import UpdateProvider


class UpdateProviderGeneric(UpdateProvider):
    """Update provider for generic shell command."""

    name: str = "generic"
    cmd: list[str] = []
    pattern: str = r"(?P<name>.+)"  # Match anything to group 'name'

    def fetch_updates(self) -> list[PackageUpdate]:
        """Get list of updates from generic shell command."""
        try:
            available_updates_run = subprocess.run(  # noqa: S603
                self.cmd,
                check=True,
                capture_output=True,
                text=True,
                shell=False,
            )
            return self.parse_updates(available_updates_run.stdout.splitlines(), self.pattern)
        except Exception as error:
            message = f"unexpected error: {error}"
            raise UpdateProviderError(message, self.name) from error

    model_config = ConfigDict(extra="forbid")



"""Test config module."""

import tomllib
from os import environ
from pathlib import Path
from unittest import mock

import pytest

from siun.config import get_config
from siun.errors import ConfigError
from siun.models import ClickColor

CONFIG_MISSING_WEIGHTS = """
[[v2_criteria]]
name = "custom"
"""

CONFIG_CUSTOM_STATE_DIR = """
state_dir = "/tmp/siun-test-state"
"""

CONFIG_LEGACY_THRESHOLDS = """
thresholds = { available = 1, warning = 2, critical = 3 }
[[update_providers]]
name = "pacman"
[[v2_criteria]]
name = "count"
weight = 1
count = 15
"""

CONFIG_LEGACY_CRITERIA = """
[[update_providers]]
name = "pacman"
[criteria]
critical_pattern = "^package$"
critical_weight = 1
count_threshold = 30
count_weight = 1
"""

CONFIG_V2_THRESHOLDS = """
[[update_providers]]
name = "pacman"
[[v2_thresholds]]
name = "critical"
score = 3
color = "red"
text = "cr"

[[v2_thresholds]]
name = "warning"
score = 2
color = "yellow"
text = "wa"

[[v2_thresholds]]
name = "available"
score = 1
color = "green"
text = "av"

[[v2_criteria]]
name = "count"
weight = 1
count = 15
"""

CONFIG_W_DUPLICATE_T_NAMES = """
[[update_providers]]
name = "pacman"
[[v2_thresholds]]
name = "dupe"
score = 3
text = "cr"

[[v2_thresholds]]
name = "dupe"
score = 2
text = "wa"

[[v2_criteria]]
name = "count"
weight = 1
count = 15
"""

CONFIG_W_INVALID_NOTIFICATION_THRESHOLD = """
[[update_providers]]
name = "pacman"
[[v2_criteria]]
name = "count"
weight = 1
count = 15
[notification]
threshold = "non-existent-threshold"
"""

CONFIG_W_NEWS = """
[[news]]
url = "https://pypi.org/rss/project/siun/releases.xml"
title = "Test Title"
max_items = 1
"""

# Override `$HOME` for consistent tests
environ["HOME"] = "/tmp/siun-tests"  # noqa: S108


def mock_read_config(_):
    """Mock empty config file."""
    return tomllib.loads("")


@pytest.mark.usefixtures("patch_is_path_world_writable")
class TestConfig:
    """Test Config class."""

    @mock.patch("siun.config._read_config", mock_read_config)
    def test_default_config(self, default_config, default_update_providers):
        """Test empty user config."""
        with (
            mock.patch("pathlib.Path.exists", return_value=True),
            mock.patch("pathlib.Path.is_file", return_value=True),
        ):
            config = get_config()

        assert config == default_config
        assert config.update_providers == default_update_providers

    @mock.patch("siun.config._read_config", return_value=tomllib.loads(CONFIG_MISSING_WEIGHTS))
    def test_missing_weights(self, mock_read_config):
        """Test user config with missing weights."""
        with mock.patch("siun.config.get_default_config_dir"), pytest.raises(ConfigError) as exc_info:
            get_config()

        mock_read_config.assert_called_once()
        assert "'v2_criteria.0.weight': Field required" in str(exc_info.value)

    @mock.patch("siun.config._read_config", return_value=tomllib.loads(CONFIG_CUSTOM_STATE_DIR))
    def test_custom_state_dir(self, mock_read_config, default_config):
        """Test custom state file path."""
        with mock.patch("siun.config.get_default_config_dir"):
            config = get_config()
        assert config != default_config
        assert config.state_dir == Path("/tmp/siun-test-state")  # noqa: S108

        mock_read_config.assert_called_once()

    @mock.patch("siun.config._read_config", return_value=tomllib.loads(CONFIG_W_INVALID_NOTIFICATION_THRESHOLD))
    def test_invalid_notification_threshold(self, mock_read_config, default_config):
        """Test notification.threshold validation."""
        with mock.patch("siun.config.get_default_config_dir"), pytest.raises(ConfigError) as exc_info:  # noqa: PT011
            get_config()

        mock_read_config.assert_called_once()
        assert "notification.threshold must be one of" in str(exc_info.value)

    @mock.patch("siun.config._read_config", mock_read_config)
    def test_xdg_state_home_set(self, default_config):
        """Test XDG_STATE_HOME being set."""
        with (
            mock.patch.dict(environ, clear=True),
            mock.patch("pathlib.Path.exists", return_value=True),
            mock.patch("pathlib.Path.is_file", return_value=True),
        ):
            environ["XDG_STATE_HOME"] = "/tmp/siun-tests/state"  # noqa: S108
            config = get_config()

        assert config.state_dir == Path("/tmp/siun-tests/state/siun")  # noqa: S108

    @mock.patch("siun.config._read_config", side_effect=OSError)
    def test_os_error(self, default_config):
        """Test handling of OSError."""
        with mock.patch("siun.config.get_default_config_dir"), pytest.raises(ConfigError) as exc_info:  # noqa: PT011
            get_config()

        assert "failed to open config file" in str(exc_info.value)

    @mock.patch("siun.config._read_config", side_effect=tomllib.TOMLDecodeError)
    def test_toml_error(self, default_config):
        """Test handling of TOMLDecodeError."""
        with mock.patch("siun.config.get_default_config_dir"), pytest.raises(ConfigError) as exc_info:  # noqa: PT011
            get_config()

        assert "config file not valid" in str(exc_info.value)

    @mock.patch("siun.config._read_config", return_value=tomllib.loads(CONFIG_W_NEWS))
    def test_config_with_news_source(self, default_config):
        """Test config with news source."""
        with (
            mock.patch("pathlib.Path.exists", return_value=True),
            mock.patch("pathlib.Path.is_file", return_value=True),
        ):
            config = get_config()

        assert len(config.news) == 1
        news_source = config.news[0]
        assert news_source.url == "https://pypi.org/rss/project/siun/releases.xml"
        assert news_source.title == "Test Title"
        assert news_source.max_items == 1


@pytest.mark.usefixtures("patch_is_path_world_writable")
class TestThresholdsConfig:
    """Test config with v2 thresholds."""

    @mock.patch("siun.config._read_config", return_value=tomllib.loads(CONFIG_LEGACY_THRESHOLDS))
    def test_legacy_thresholds_raises_error(self, mock_read_config):
        """Test config using legacy 'thresholds' dict."""
        with mock.patch("siun.config.get_default_config_dir"), pytest.raises(ConfigError) as exc_info:
            get_config()

        mock_read_config.assert_called_once()
        assert "deprecated config fields: \n- 'thresholds'" in str(exc_info.value)

    @mock.patch("siun.config._read_config", return_value=tomllib.loads(CONFIG_V2_THRESHOLDS))
    def test_v2_thresholds(self, mock_read_config):
        """Test config using v2_thresholds list."""
        with (
            mock.patch("siun.config.get_default_config_dir"),
        ):
            config = get_config()

        names = [t.name for t in config.v2_thresholds]
        scores = [t.score for t in config.v2_thresholds]
        colors = [getattr(t, "color", None) for t in config.v2_thresholds]

        assert set(names) == {"critical", "warning", "available"}
        assert set(scores) == {1, 2, 3}
        assert set(colors) == {ClickColor.red, ClickColor.yellow, ClickColor.green}
        mock_read_config.assert_called_once()

    @mock.patch("siun.config._read_config", return_value=tomllib.loads(CONFIG_W_DUPLICATE_T_NAMES))
    def test_v2_thresholds_name_uniqueness(self, mock_read_config):
        """Test v2_thresholds require unique names."""
        with (
            mock.patch("siun.config.get_default_config_dir"),
            pytest.raises(ConfigError) as exc_info,
        ):
            get_config()

        mock_read_config.assert_called_once()
        assert "threshold must have a unique name" in str(exc_info.value)


@pytest.mark.usefixtures("patch_is_path_world_writable")
class TestCriteriaConfig:
    """Test config with v2 criteria."""

    @mock.patch("siun.config._read_config", return_value=tomllib.loads(CONFIG_LEGACY_CRITERIA))
    def test_legacy_criteria_raises_error(self, mock_read_config):
        """Test config using legacy 'criteria' dict."""
        with (
            mock.patch("siun.config.get_default_config_dir"),
            pytest.raises(ConfigError) as exc_info,
        ):
            get_config()

        mock_read_config.assert_called_once()
        assert "deprecated config fields: \n- 'criteria'" in str(exc_info.value)



"""Module for feeds and news."""

import calendar
import datetime
import json
from pathlib import Path

import click

from siun.errors import (
    SiunCLIError,
)
from siun.models import ClickColor, NewsEntry, NewsProvider
from siun.util import safely_write_to_disk

INSTALLED_FEATURES: set[str] = set()

try:
    import feedparser  # noqa: F401

    INSTALLED_FEATURES.add("news")
except ImportError:
    pass


def parse_feed_entries(source: NewsProvider) -> tuple[str, list[NewsEntry]]:
    """Parse feed entries from a news source."""
    parsed_feed = feedparser.parse(source.url, etag=source.etag, modified=source.last_modified)
    # Update ETag and Last-Modified for future requests
    source._etag = getattr(parsed_feed, "etag", None)
    source._last_modified = getattr(parsed_feed, "modified", None)

    if not source.title:
        source.title = parsed_feed.feed.get("title", "No title")
    entries: list[NewsEntry] = []
    for entry in parsed_feed.entries[: source.max_items] or []:
        entry_datetime = datetime.datetime.fromtimestamp(calendar.timegm(entry.published_parsed), tz=datetime.UTC)
        entries.append(
            NewsEntry(
                title=entry.get("title", "Untitled"),
                link=entry.get("link", ""),
                published_at=entry_datetime.isoformat(),
            )
        )
    return source.title, entries


def load_news_state(sources: list[NewsProvider], last_update_path: Path) -> None:
    """Load ETag and Last-Modified headers from disk for news sources."""
    from_disk: list[dict[str, str]] = []
    if last_update_path.exists() and last_update_path.is_file():
        with last_update_path.open("r", encoding="utf-8") as file_obj:
            try:
                from_disk = json.loads(file_obj.read())
            except json.JSONDecodeError as error:
                raise SiunCLIError(
                    message=f"Failed to parse last news update state from disk; state path: {last_update_path}"
                ) from error

    for source in sources:
        for saved_source in from_disk:
            if source.url == saved_source.get("url", ""):
                source._etag = saved_source.get("etag", None)
                source._last_modified = saved_source.get("last_modified", None)
                if not source.title:
                    source.title = saved_source.get("title", "No title")
                break


def save_news_state(sources: list[NewsProvider], last_update_path: Path) -> None:
    """Write news state to disk."""
    return safely_write_to_disk(
        content=json.dumps(
            [source.model_dump(include={"url", "etag", "last_modified", "title"}) for source in sources]
        ),
        target_path=last_update_path,
    )


def format_news_entries(news_entries: dict[str, list[NewsEntry]], nocolor: bool) -> str:
    """Format news entries for output."""
    output = []
    for feed_title, entries in news_entries.items():
        output.append(click.style(f"{feed_title}:\n", fg=ClickColor.yellow.value if not nocolor else None))
        for entry in entries:
            output.append(f"- {entry.title}")
            output.append(click.style(f"  {entry.link}", fg=ClickColor.blue.value if not nocolor else None))
            output.append(f"  {entry.published_at}\n")
        if not entries:
            output.append("No new entries.\n")
    return "\n".join(output)



"""Criteria module."""

import re
from typing import Any


class SiunCriterion:
    """Base class for criteria."""

    def is_fulfilled(self, criteria_settings: dict[str, Any], available_updates: list[str]) -> bool:
        """Override me."""
        raise NotImplementedError


class CriterionAvailable(SiunCriterion):
    """Check if there are any available updates."""

    def is_fulfilled(self, criteria_settings: dict[str, Any], available_updates: list[str]) -> bool:
        """Check criterion."""
        return bool(available_updates)


class CriterionCount(SiunCriterion):
    """Check if count of available updates has exceeded threshold."""

    def is_fulfilled(self, criteria_settings: dict[str, Any], available_updates: list[str]) -> bool:
        """Check criterion."""
        is_exceeded: bool = len(available_updates) >= criteria_settings["count"]
        return is_exceeded


class CriterionPattern(SiunCriterion):
    """Check if list of available updates contains updates according to pattern."""

    def is_fulfilled(self, criteria_settings: dict[str, Any], available_updates: list[str]) -> bool:
        """Check criterion."""
        regex = re.compile(criteria_settings["pattern"])
        matches = list(filter(regex.match, available_updates))

        return bool(matches)



"""Custom error module."""

from pathlib import Path
from typing import final

from click import ClickException


@final
class UpdateProviderError(Exception):
    """Wrap Exception for update provider errors."""

    def __init__(self, message: str, provider_name: str):
        super().__init__(message)
        self.message = f"[{provider_name}] {message}"
        self.provider_name = provider_name


@final
class CriterionError(Exception):
    """Wrap Exception for criterion errors."""

    def __init__(self, message: str, criterion_name: str | None):
        super().__init__(message)
        self.message = message
        self.criterion_name = criterion_name


@final
class ConfigError(Exception):
    """Custom error type for configuration issues."""

    def __init__(self, message: str, config_path: Path):
        super().__init__(message)
        self.message = message
        self.config_path = config_path


@final
class SiunStateUpdateError(Exception):
    """Wrap Exception for errors on updating the internal state."""

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


@final
class SiunGetUpdatesError(Exception):
    """Wrap Exception for errors on getting available updates."""

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


@final
class SiunNotificationError(Exception):
    """Wrap Exception for errors related to notifications."""

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class SiunCLIError(ClickException):
    """Wrap ClickException for explicit abort on error."""



"""Flatpak package update provider."""

from __future__ import annotations

import subprocess

from pydantic import ConfigDict

from siun.errors import UpdateProviderError
from siun.models import PackageUpdate
from siun.providers.base import UpdateProvider


class UpdateProviderFlatpak(UpdateProvider):
    """Update provider for flatpak."""

    name: str = "flatpak"
    list_apps: bool = True
    list_runtimes: bool = True
    _default_cmds: list[list[str]] = [["flatpak", "remote-ls", "--updates", "--columns=name,version,branch,commit"]]

    def fetch_updates(self) -> list[PackageUpdate]:
        """Get list of updates from flatpak."""
        cmd = self.pick_cmd(self._default_cmds)
        if not self.list_apps:
            cmd.append("--runtime")
        if not self.list_runtimes:
            cmd.append("--app")

        try:
            available_updates_run = subprocess.run(  # noqa: S603
                cmd,
                check=True,
                capture_output=True,
                text=True,
                shell=False,
            )
            return self.parse_updates(available_updates_run.stdout.splitlines(), "")

        except Exception as error:
            message = f"unexpected error: {error}"
            raise UpdateProviderError(message, self.name) from error

    def parse_updates(self, lines: list[str], pattern: str) -> list[PackageUpdate]:
        """Parse list of available update strings into list of PackageUpdate objects."""
        available_updates: list[PackageUpdate] = []
        for line in lines:
            fields = line.split("\t")
            if len(fields) != 4:
                message = f"failed to parse output: {line}"
                raise UpdateProviderError(message, self.name)

            name, version, branch, commit = fields
            available_updates.append(
                PackageUpdate(
                    name=name,
                    old_version=version or branch or commit,
                    new_version=None,
                    provider=self.name,
                )
            )
        return available_updates

    model_config = ConfigDict(extra="forbid")



from .criteria import CRITERION_REGISTRY as CRITERION_REGISTRY
from .criteria import CriterionArchaudit as CriterionArchaudit
from .criteria import CriterionAvailable as CriterionAvailable
from .criteria import CriterionCount as CriterionCount
from .criteria import CriterionCustom as CriterionCustom
from .criteria import CriterionPattern as CriterionPattern
from .criteria import V2Criterion as V2Criterion
from .formatting import ClickColor as ClickColor
from .formatting import FormatObject as FormatObject
from .news import NewsEntry as NewsEntry
from .news import NewsProvider as NewsProvider
from .thresholds import V2Threshold as V2Threshold
from .updates import PackageUpdate as PackageUpdate
from .updates import Updates as Updates



"""Module for models related to news feeds."""

from __future__ import annotations

from pydantic import BaseModel, computed_field


class NewsProvider(BaseModel):
    """Struct for news sources."""

    url: str
    title: str = ""
    max_items: int = 3
    _etag: str | None = None
    _last_modified: str | None = None

    @computed_field
    @property
    def etag(self) -> str | None:
        """Get ETag if available."""
        if getattr(self, "_etag", False):
            return self._etag

        return None

    @computed_field
    @property
    def last_modified(self) -> str | None:
        """Get last_modified if available."""
        if getattr(self, "_last_modified", False):
            return self._last_modified

        return None


class NewsEntry(BaseModel):
    """Struct for news entries."""

    title: str
    link: str
    published_at: str



"""Test state module."""

import io
import stat
from os import environ
from pathlib import Path
from unittest import mock

import pytest

from siun.errors import CriterionError
from siun.models import CriterionCustom, FormatObject, PackageUpdate, Updates
from siun.state import BUILTIN_CRITERIA, _load_user_criteria, get_merged_criteria, load_state
from siun.util import get_default_criteria_dir


class TestUpdates:
    """Test Updates class."""

    def test_defaults_ok(self, default_config, default_thresholds):
        """Test no available updates."""
        updates = Updates(thresholds=default_thresholds, criteria_settings=default_config.v2_criteria)
        updates.evaluate(criteria=BUILTIN_CRITERIA, available_updates=[])
        result = updates.text_value

        assert result == "No matches."

    def test_defaults_available(self, default_config, default_thresholds, updates_single):
        """Test available updates."""
        updates = Updates(thresholds=default_thresholds, criteria_settings=default_config.v2_criteria)
        updates.evaluate(criteria=BUILTIN_CRITERIA, available_updates=[updates_single])
        result = updates.text_value

        assert result == "Updates available"

    def test_defaults_recommended(self, default_config, default_thresholds):
        """Test recommended updates."""
        updates = Updates(thresholds=default_thresholds, criteria_settings=default_config.v2_criteria)
        updates.evaluate(
            criteria=BUILTIN_CRITERIA,
            available_updates=[
                PackageUpdate(name="siun", provider="pacman"),
                PackageUpdate(name="linux", provider="pacman"),
            ],
        )
        result = updates.text_value

        assert result == "Updates recommended"

    def test_defaults_required(self, default_config, default_thresholds):
        """Test required updates."""
        updates = Updates(thresholds=default_thresholds, criteria_settings=default_config.v2_criteria)
        updates.evaluate(
            criteria=BUILTIN_CRITERIA,
            available_updates=[
                PackageUpdate(name="siun", provider="pacman"),
                PackageUpdate(name="linux", provider="pacman"),
                *[PackageUpdate(name="package", provider="pacman")] * 15,
            ],
        )
        result = updates.text_value

        assert result == "Updates required"

    @mock.patch("siun.state.Path.open")
    def test_read_state(self, mock_open):
        """Test reading existing state."""
        json_content = io.StringIO(
            "{"
            '"last_update": "1970-01-01T01:00:00Z", '
            '"state": "OK", '
            '"matched_criteria": {}, "available_updates": [{"name": "siun", "provider": "pacman"}]'
            "}"
        )
        mock_open.return_value = json_content
        mock_file = mock.MagicMock()
        updates = load_state(mock_file)

        assert updates
        assert updates.available_updates == [PackageUpdate(name="siun", provider="pacman")]

    @mock.patch("siun.state.Path.open")
    def test_read_state_handles_deprecated_types(self, mock_open):
        """Test loading state from disk handles custom types."""
        # `state` is deprecated
        json_content = io.StringIO(
            "{"
            '"last_update": "1970-01-01T01:00:00Z", '
            '"state": "OK", "thresholds": [], '
            '"matched_criteria": {}, "available_updates": []'
            "}"
        )
        mock_open.return_value = json_content
        mock_file = mock.MagicMock()
        updates = load_state(mock_file)

        assert updates
        assert updates.available_updates == []
        assert updates.match is None

    @mock.patch("siun.state.Path.open")
    def test_read_state_handles_deprecated_custom_types(self, mock_open):
        """Test loading state from disk handles custom types."""
        # `py-type` struct is deprecated
        json_content = io.StringIO(
            "{"
            '"last_update": "1970-01-01T01:00:00Z", '
            '"state": {"py-type": "State", "value": "OK"}, "thresholds": [], '
            '"matched_criteria": {}, "available_updates": []'
            "}"
        )
        mock_open.return_value = json_content
        mock_file = mock.MagicMock()
        updates = load_state(mock_file)

        assert updates.match is None  # Deprecated type is treated like unknown state

    def test_load_state_file_missing(self, tmp_path):
        """Test loading state if state file does not exist."""
        state_file = tmp_path / "siun-missing.json"
        result = load_state(state_file)

        assert result is None

    @mock.patch("siun.state._load_user_criteria", side_effect=RuntimeError("fail!"))
    def test_update_raises_criterion_error_on_user_criteria_failure(
        self,
        mock_load_user_criteria,
        default_config,
        default_thresholds,
        updates_single,
    ):
        """Test get_merged_criteria raises CriterionError if user criteria loading fails."""
        with pytest.raises(CriterionError) as excinfo:
            get_merged_criteria(criteria_settings=default_config.v2_criteria)

        assert "Error loading criteria: fail!" in str(excinfo.value)

    def test_format_object_populated(self, default_config, default_thresholds):
        """Test format_object returns correct values for populated state."""
        updates = Updates(thresholds=default_thresholds, criteria_settings=default_config.v2_criteria)
        updates.available_updates = [
            PackageUpdate(name="siun", provider="pacman"),
            PackageUpdate(name="linux", provider="pacman"),
        ]
        updates.matched_criteria = {
            "available": {"weight": 1, "name_short": "av"},
            "count": {"weight": 2, "name_short": "co"},
        }
        updates.match = default_thresholds[1]  # Assume at least two thresholds
        fmt = updates.format_object
        assert isinstance(fmt, FormatObject)
        assert fmt.available_updates == "siun, linux"
        assert fmt.matched_criteria == "available, count"
        assert fmt.matched_criteria_short == "av,co"
        assert fmt.score == 3
        assert fmt.status_text == updates.text_value
        assert fmt.update_count == 2

    def test_match_reset_when_no_thresholds_matched(self, default_config, default_thresholds):
        """Test that match is reset to None when score is below all thresholds."""
        updates = Updates(thresholds=default_thresholds, criteria_settings=default_config.v2_criteria)
        updates.match = default_thresholds[-1]
        updates.evaluate(criteria=BUILTIN_CRITERIA, available_updates=[])
        assert updates.match is None

    @pytest.mark.parametrize("updates_multiple", [2], indirect=True)
    def test_match_not_reset_if_threshold_matched(self, default_config, default_thresholds, updates_multiple):
        """Test that match is set if a threshold is matched."""
        updates = Updates(thresholds=default_thresholds, criteria_settings=default_config.v2_criteria)
        updates.evaluate(criteria=BUILTIN_CRITERIA, available_updates=updates_multiple)
        assert updates.match is not None
        assert updates.match.score <= updates.score

    @pytest.mark.parametrize("updates_multiple", [3], indirect=True)
    def test_match_reset_on_score_drop(self, default_config, default_thresholds, updates_multiple):
        """Test that match is reset if score drops below all thresholds after being previously set."""
        updates = Updates(thresholds=default_thresholds, criteria_settings=default_config.v2_criteria)
        updates.evaluate(criteria=BUILTIN_CRITERIA, available_updates=updates_multiple)
        assert updates.match is not None
        updates.evaluate(criteria=BUILTIN_CRITERIA, available_updates=[])
        assert updates.score == 0
        assert updates.match is None


class TestCustomCriteria:
    """Test custom criteria."""

    def test_load_custom_criterion(self, tmp_path):
        """Test custom criteria can be loaded."""
        criteria_settings = [CriterionCustom(name="test_criterion", weight=2)]
        include_path = tmp_path / "criteria"
        include_path.mkdir()
        criterion_content = """class SiunCriterion:
    def is_fulfilled(self, criteria_settings, available_updates):
        return True
        """
        with Path.open(include_path / Path("test_criterion.py"), "w+") as criterion_file:
            criterion_file.write(criterion_content)
        user_criteria = _load_user_criteria(criteria_settings=criteria_settings, include_path=include_path)
        assert user_criteria
        assert "test_criterion" in user_criteria

    @pytest.mark.parametrize("criteria_settings", [[CriterionCustom(name="test_criterion", weight=0)], []])
    def test_custom_criterion_not_loaded_wo_weight(self, tmp_path, criteria_settings):
        """Test custom criteria not getting loaded without a configured weight."""
        include_path = tmp_path / "criteria"
        include_path.mkdir()
        criterion_content = """class SiunCriterion:
    def is_fullfilled(self, criteria_settings, available_updates):
        return True
        """
        with Path.open(include_path / Path("test_criterion.py"), "w+") as criterion_file:
            criterion_file.write(criterion_content)
        user_criteria = _load_user_criteria(criteria_settings=criteria_settings, include_path=include_path)
        assert isinstance(user_criteria, dict)
        assert "test_criterion" not in user_criteria

    def test__default_criteria_dir(self):
        """Test get_default_criteria_dir with XDG_CONFIG_HOME set."""
        with mock.patch.dict(environ, clear=True):
            environ["XDG_CONFIG_HOME"] = "/tmp/siun-tests/config"  # noqa: S108
            assert get_default_criteria_dir() == Path("/tmp/siun-tests/config/siun/criteria")  # noqa: S108

    def test__default_criteria_dir_wo_config_home(self):
        """Test get_default_criteria_dir without XDG_CONFIG_HOME set."""
        with mock.patch.dict(environ, clear=True):
            environ["HOME"] = "/tmp/siun-tests/no_config_home"  # noqa: S108
            assert get_default_criteria_dir() == Path("/tmp/siun-tests/no_config_home/.config/siun/criteria")  # noqa: S108

    def test_load_user_criteria_world_writable_dir(self, tmp_path):
        """Test _load_user_criteria exception if criteria dir is world-writable."""
        criteria_settings = [CriterionCustom(name="test_criterion", weight=1)]
        include_path = tmp_path / "criteria"
        include_path.mkdir()
        assert include_path.exists()
        assert include_path.is_dir()

        with mock.patch("pathlib.Path.stat") as mock_stat:

            class DummyStat:
                # Mock st_mode to be world-writable and a directory
                st_mode = stat.S_IWOTH | stat.S_IFDIR

            mock_stat.return_value = DummyStat()
            assert include_path.exists()
            assert include_path.is_dir()
            with pytest.raises(ImportError) as excinfo:
                _load_user_criteria(criteria_settings=criteria_settings, include_path=include_path)
            assert "world-writable" in str(excinfo.value)



"""Module for checking for updates and evaluating criteria."""

import datetime
from pathlib import Path

from siun.errors import (
    SiunGetUpdatesError,
    SiunStateUpdateError,
)
from siun.models import V2Criterion, V2Threshold
from siun.models.updates import Updates
from siun.providers import UpdateProvider
from siun.state import get_merged_criteria, get_package_updates, load_state


def _is_cache_stale(existing_state, now, min_age):
    return existing_state and existing_state.last_update < (now - min_age)


def _evaluate_state(state, criteria_dict, updates):
    try:
        state.evaluate(criteria_dict, available_updates=updates)
    except SiunStateUpdateError as error:
        raise SiunGetUpdatesError(error.message) from error


def _persist_state(state, path):
    try:
        state.persist_state(path)
    except Exception as error:
        message = f"failed to write state to disk: {error}"
        raise SiunGetUpdatesError(message) from error


def get_updates(
    *,
    no_cache: bool,
    no_update: bool,
    criteria: list[V2Criterion],
    thresholds: list[V2Threshold],
    cache_min_age_minutes: int,
    state_file_path: Path,
    update_providers: list[UpdateProvider],
) -> Updates:
    """Get available updates and evaluate criteria."""
    now = datetime.datetime.now(tz=datetime.UTC)
    cache_min_age = datetime.timedelta(minutes=cache_min_age_minutes)
    criteria_dict = get_merged_criteria(criteria_settings=criteria)

    if no_cache:
        state = Updates(criteria_settings=criteria, thresholds=thresholds)
        if no_update:
            return state
        _evaluate_state(state, criteria_dict, get_package_updates(update_providers))
        return state

    try:
        existing_state = load_state(state_file_path)
    except Exception as error:
        message = f"failed to load state from disk: {error}"
        raise SiunGetUpdatesError(message) from error

    should_update_cache = not existing_state or _is_cache_stale(existing_state, now, cache_min_age)
    needs_update = False

    if existing_state:
        state = existing_state
        state.last_match = state.match
        state.thresholds = thresholds
        state.criteria_settings = criteria
        _evaluate_state(state, criteria_dict, existing_state.available_updates)
        if state.last_match != state.match:
            needs_update = True
    else:
        state = Updates(criteria_settings=criteria, thresholds=thresholds)

    if no_update:
        return state

    if should_update_cache:
        _evaluate_state(state, criteria_dict, get_package_updates(update_providers))
        _persist_state(state, state_file_path)
    elif needs_update:
        _persist_state(state, state_file_path)

    return state



"""Module for desktop notifications."""

from enum import Enum
from string import Template
from typing import TYPE_CHECKING, cast

from pydantic import BaseModel, ConfigDict, Field, field_validator

if TYPE_CHECKING:
    from siun.models import FormatObject


# NOTE: The reason for all this INSTALLED_FEATURES logic is to provide a nice
# error message if a user tries to configure the notification without
# installing the required dependencies.
INSTALLED_FEATURES: set[str] = set()

try:
    from dbus import Byte as DBusByte
    from dbus import Interface as DBusInterface
    from dbus import SessionBus

    INSTALLED_FEATURES.add("notification")
except ImportError:
    DBusByte = int  # dbus.Byte is a subtype of int, so this will pass Config validation

# Typing alias for dbus.Byte, so the correct type can be inferred later
NotificationByte: type = cast(type, DBusByte)


class NotificationUrgency(Enum):
    """Urgency levels for notifications."""

    low = NotificationByte(0)
    normal = NotificationByte(1)
    critical = NotificationByte(2)


class UpdateNotification(BaseModel):
    """Desktop notification struct."""

    app_name: str = "siun"
    notification_id: int = 0
    icon: str = Field(default="siun-icon")
    title: str = Field(default="$status_text")
    message: str = Field(default="$available_updates")
    actions: list = []
    hints: dict = {}
    timeout: int = Field(default=5000)
    urgency: NotificationUrgency | None = None
    threshold: str  # NOTE: This field gets validated in the Config model

    model_config = ConfigDict(extra="forbid")

    @field_validator("urgency", mode="before")
    def urgency_must_be_enum(
        cls,
        value: str | None,
    ) -> NotificationUrgency | None:
        """Make sure urgency is a valid NotificationUrgency."""
        if value is None:
            return value

        try:
            urgency: NotificationUrgency = NotificationUrgency[value].value
        except KeyError as err:
            message = f"input should be a valid urgency (low|normal|critical), unable to parse '{value}' as urgency"
            raise ValueError(message) from err
        return urgency

    def show(self):
        """Show notification."""
        session_bus = SessionBus()
        notify = session_bus.get_object("org.freedesktop.Notifications", "/org/freedesktop/Notifications")
        notify_interface = DBusInterface(notify, "org.freedesktop.Notifications")

        notify_interface.Notify(
            self.app_name,
            self.notification_id,
            self.icon,
            self.title,
            self.message,
            self.actions,
            self.hints,
            self.timeout,
        )

    def fill_templates(self, format_object: "FormatObject"):
        """Fill template strings with format variables."""
        title_template = Template(self.title)
        self.title = title_template.safe_substitute(**format_object.model_dump())
        message_template = Template(self.message)
        self.message = message_template.safe_substitute(**format_object.model_dump())



"""Test update providers module."""

from unittest import mock

import pytest
from pydantic import ValidationError

from siun.errors import UpdateProviderError
from siun.models import PackageUpdate
from siun.providers import UpdateProvider, UpdateProviderGeneric, UpdateProviderPacman


class TestUpdateProviderBase:
    """Test base interface for UpdateProviders."""

    def test_fetch_updates(self):
        """Test fetch_updates."""
        provider = UpdateProvider(name="dummy")
        with pytest.raises(NotImplementedError):
            provider.fetch_updates()

    def test_parse_updates(self):
        """Test parse_updates success."""
        provider = UpdateProvider(name="dummy")
        pattern = r"(?P<name>.+)"
        lines = ["package-x", "another package 1.2.3"]

        parsed_updates = provider.parse_updates(lines, pattern)
        names = [update.name for update in parsed_updates]
        assert "package-x" in names
        assert "another package 1.2.3" in names

    def test_parse_updates_empty(self):
        """Test parse_updates success if there are no updates."""
        provider = UpdateProvider(name="dummy")
        pattern = r"(?P<name>.+)"
        lines = []

        parsed_updates = provider.parse_updates(lines, pattern)
        assert parsed_updates == []

    def test_parse_updates_fails_on_no_matches(self):
        """Test parse_updates fails if output doesn't match pattern."""
        provider = UpdateProvider(name="dummy")
        pattern = r"(?P<name>a-z+)"
        lines = ["package-0", "package 1.2.3"]

        with pytest.raises(UpdateProviderError) as excinfo:
            provider.parse_updates(lines, pattern)
        assert "failed to parse output: package-0" in str(excinfo.value)


class TestUpdateProviderPacman:
    """Test UpdateProviderPacman."""

    @mock.patch("siun.providers.pacman.UpdateProviderPacman.pick_cmd", return_value=[":"])
    @mock.patch("subprocess.run")
    def test_fetch_updates(self, mock_run, mock_pick_cmd):
        """Test fetch_updates success."""
        mock_run.return_value = mock.Mock(stdout="siun 1.0.0 -> 2.0.0", returncode=0)
        provider = UpdateProviderPacman()
        available_updates = provider.fetch_updates()
        assert available_updates == [
            PackageUpdate(name="siun", old_version="1.0.0", new_version="2.0.0", provider="pacman")
        ]

    @mock.patch("siun.providers.pacman.UpdateProviderPacman.pick_cmd", return_value=[":"])
    @mock.patch("subprocess.run", side_effect=PermissionError("Permission denied"))
    def test_fetch_updates_cmd_fails(self, mock_run, mock_pick_cmd):
        """Test fetch_updates with failing cmd."""
        provider = UpdateProviderPacman()
        with pytest.raises(UpdateProviderError) as excinfo:
            provider.fetch_updates()
        assert "Permission denied" in str(excinfo.value)

    @mock.patch("siun.providers.pacman.UpdateProviderPacman.pick_cmd", return_value=[":"])
    def test_fetch_updates_cmd_not_found(self, fp):
        """Test fetch_updates with cmd not installed."""
        provider = UpdateProviderPacman()
        with pytest.raises(UpdateProviderError) as excinfo:
            provider.fetch_updates()
        assert "No such file or directory: ':'" in str(excinfo.value)

    def test_fetch_updates_invalid_cmd(self, fp):
        """Test fetch_updates with invalid cmd."""
        fp.register([":"], stdout="")

        with pytest.raises(ValidationError):
            UpdateProviderPacman(cmd="this_should_be_a_list.sh")


class TestUpdateProviderGeneric:
    """Test UpdateProviderGeneric."""

    def test_custom_pattern(self, fp):
        """Test parse_updates with custom pattern."""
        fp.register([":"], stdout="siun 2.7.18")

        provider = UpdateProviderGeneric(cmd=[":"], pattern=r"(?P<name>[a-z]+)\s+(?P<new_version>[0-9\.]+)")
        available_updates = provider.fetch_updates()
        assert available_updates == [
            PackageUpdate(name="siun", new_version="2.7.18", old_version=None, provider="generic")
        ]



"""Test main function and CLI."""

import datetime
from pathlib import Path
from unittest import mock

import pytest

from siun.check import get_updates
from siun.cli import _handle_notification
from siun.errors import SiunNotificationError
from siun.models import CriterionAvailable, CriterionCount
from siun.providers import UpdateProvider
from siun.state import Updates


class TestMain:
    """Test main function."""

    @mock.patch("siun.cli.INSTALLED_FEATURES", set())
    def test__handle_notification_missing_feature(self, notification_mock):
        """Test _handle_notification raises SiunNotificationError if feature missing."""
        notification = notification_mock("warning")
        config = mock.Mock()
        config.notification = notification
        config.mapped_thresholds = {"warning": mock.Mock(score=10)}
        state = mock.Mock()
        state.match = mock.Mock(score=15)
        state.last_match = None
        state.format_object = {}

        with pytest.raises(SiunNotificationError) as excinfo:
            _handle_notification(config, state)
        assert "notifications require the 'notification' feature" in str(excinfo.value)

    @mock.patch("siun.cli.INSTALLED_FEATURES", {"notification"})
    def test__handle_notification_threshold_logic(self, notification_mock):
        """Test _handle_notification does not show notification if match.score <= threshold_score."""
        notification = notification_mock("warning")
        config = mock.Mock()
        config.notification = notification
        config.mapped_thresholds = {"warning": mock.Mock(score=10)}
        state = mock.Mock()
        state.match = mock.Mock(score=8)
        state.last_match = None
        state.format_object = {}

        _handle_notification(config, state)
        notification.show.assert_not_called()

    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state")
    def test_existing_state_sets_required_fields(
        self, mock_load_state, mock_persist_state, v2_config_w_custom_format, updates_single
    ):
        """Test loaded state receives required values from config."""
        loaded_state = Updates(
            criteria_settings=[],
            thresholds=[],
            available_updates=[updates_single],
            matched_criteria={},
            last_update=datetime.datetime.now(tz=datetime.UTC),
        )
        mock_load_state.return_value = loaded_state

        config_criteria = [
            CriterionAvailable(name="available", weight=1),
            CriterionCount(name="count", weight=2, count=1),
        ]

        result = get_updates(
            no_cache=False,
            no_update=True,
            update_providers=[UpdateProvider(name="dummy")],
            criteria=config_criteria,
            thresholds=v2_config_w_custom_format.v2_thresholds,
            cache_min_age_minutes=10,
            state_file_path=Path("/tmp/siun-test-state.json"),  # noqa: S108
        )

        assert result.criteria_settings == config_criteria
        assert result.thresholds == v2_config_w_custom_format.v2_thresholds
        assert result.available_updates == [updates_single]

    @mock.patch("siun.check.get_package_updates", return_value=[])
    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state")
    def test_get_updates_persist_state_on_match_change(
        self, mock_load_state, mock_persist_state, mock_get_package_updates, default_thresholds
    ):
        """Test persist_state called when last_match != match."""
        threshold1 = default_thresholds[0]
        threshold2 = default_thresholds[-1]

        def evaluate_side_effect(self, criteria, available_updates):
            self.match = threshold2  # new match

        with mock.patch("siun.state.Updates.evaluate", evaluate_side_effect):
            state = Updates(
                criteria_settings=[],
                thresholds=[threshold1, threshold2],
                available_updates=[],
                matched_criteria={},
                last_update=datetime.datetime.now(tz=datetime.UTC),
                match=threshold1,
                last_match=threshold2,
            )
            mock_load_state.return_value = state

            get_updates(
                no_cache=False,
                no_update=False,
                update_providers=[UpdateProvider(name="dummy")],
                criteria=[],
                thresholds=[threshold1, threshold2],
                cache_min_age_minutes=10,
                state_file_path=Path("/tmp/siun-test-state.json"),  # noqa: S108
            )

        mock_persist_state.assert_called_once()

    @mock.patch("siun.check.get_package_updates", return_value=[])
    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state")
    def test_get_updates_no_persist_state_when_match_unchanged(
        self, mock_load_state, mock_persist_state, mock_get_package_updates, default_thresholds
    ):
        """Test persist_state not called when last_match == match."""
        threshold = default_thresholds[0]

        def evaluate_side_effect(self, criteria, available_updates):
            pass  # match remains unchanged

        with mock.patch("siun.state.Updates.evaluate", evaluate_side_effect):
            state = Updates(
                criteria_settings=[],
                thresholds=[threshold],
                available_updates=[],
                matched_criteria={},
                last_update=datetime.datetime.now(tz=datetime.UTC),
                match=threshold,
                last_match=threshold,
            )
            mock_load_state.return_value = state

            get_updates(
                no_cache=False,
                no_update=False,
                update_providers=[UpdateProvider(name="dummy")],
                criteria=[],
                thresholds=[threshold],
                cache_min_age_minutes=10,
                state_file_path=Path("/tmp/siun-test-state.json"),  # noqa: S108
            )

        mock_persist_state.assert_not_called()



"""Module for criteria models."""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field, computed_field

CRITERION_REGISTRY: dict[str, type[V2Criterion]] = {}


class V2Criterion(BaseModel):
    """Version 2 of the Criterion struct."""

    name: str
    short_name: str | None = Field(default=None)
    weight: int

    model_config = ConfigDict(extra="allow")

    def __init_subclass__(cls, **kwargs: object) -> None:
        """Register subclass in criterion registry."""
        super.__init_subclass__(**kwargs)
        CRITERION_REGISTRY[cls.name] = cls

    @computed_field
    @property
    def name_short(self) -> str:
        """Get short name for formatters."""
        if self.short_name is None:
            return self.name[:2]

        return self.short_name


class CriterionAvailable(V2Criterion):
    """
    Criterion for available updates.

    Considered a match if any updates are available.
    """

    name: str = "available"


class CriterionCount(V2Criterion):
    """
    Criterion for number of available updates.

    Considered a match if number of available updates equals, or exceeds `count`.
    """

    name: str = "count"
    count: int

    model_config = ConfigDict(extra="forbid")


class CriterionArchaudit(V2Criterion):
    """
    Criterion for updates flagged by archaudit.

    Considered a match if archaudit flags any of the available updates, and a fix for the issue is available.
    """

    name: str = "archaudit"
    model_config = ConfigDict(extra="forbid")


class CriterionPattern(V2Criterion):
    """
    Criterion for updates matching pattern.

    Considered a match if any of the available updates match the provided pattern.
    """

    name: str = "pattern"
    pattern: str

    model_config = ConfigDict(extra="forbid")


class CriterionCustom(V2Criterion):
    """Criterion for custom criteria."""

    name: str = "custom"
    model_config = ConfigDict(extra="allow")



"""Test check command."""

import datetime
import tempfile
import tomllib
from pathlib import Path
from unittest import mock

import pytest
from click.testing import CliRunner

from siun.cli import check, get_updates
from siun.errors import ConfigError, SiunGetUpdatesError
from siun.models import CriterionAvailable, CriterionCount, PackageUpdate
from siun.state import UpdateProvider, Updates

EMPTY_STATE = Updates(
    available_updates=[],
    matched_criteria={},
    last_update=datetime.datetime.now(tz=datetime.UTC),
)

CONFIG_CUSTOM_STATE_FILE_PATH = """
state_dir = "/tmp/siun-test-state"
"""


@pytest.mark.usefixtures("patch_is_path_world_writable")
class TestCheckCommand:
    """Test check command."""

    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state")
    @mock.patch(
        "siun.providers.UpdateProviderPacman.fetch_updates",
        return_value=[],
    )
    @mock.patch("siun.cli_utils.get_config")
    def test_check_no_available_updates_no_cache(
        self, mock_get_config, mock_fetch_available_updates, mock_read_state, mock_persist_state, default_config
    ):
        """Test check CLI command with no updates and --no-cache option."""
        mock_get_config.return_value = default_config
        runner = CliRunner()
        result = runner.invoke(check, ["-n"])
        mock_read_state.assert_not_called()
        mock_persist_state.assert_not_called()
        mock_fetch_available_updates.assert_called_once()
        assert result.exit_code == 0
        assert result.output == "No matches.\n"

    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state")
    @mock.patch(
        "siun.providers.UpdateProviderPacman.fetch_updates",
        return_value=[],
    )
    @mock.patch("siun.cli_utils.get_config")
    def test_check_with_config_path_option(
        self, mock_get_config, mockfetch_available_updates, mock_read_state, mock_persist_state, default_config
    ):
        """Test --config-path CLI option."""
        mock_get_config.return_value = default_config
        runner = CliRunner()
        with tempfile.NamedTemporaryFile(mode="r") as config_path:
            result = runner.invoke(check, ["-n", "-C", config_path.name])
            mock_get_config.assert_called_once_with(Path(config_path.name))
        mock_read_state.assert_not_called()
        mock_persist_state.assert_not_called()
        mockfetch_available_updates.assert_called_once()
        assert result.exit_code == 0
        assert result.output == "No matches.\n"

    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state")
    @mock.patch(
        "siun.providers.UpdateProviderPacman.fetch_updates",
        return_value=[],
    )
    @mock.patch("siun.cli_utils.get_config")
    def test_check_no_cache_no_update_options(
        self, mock_get_config, mockfetch_available_updates, mock_read_state, mock_persist_state, default_config
    ):
        """Test check CLI command with --no-cache and --no-update options."""
        mock_get_config.return_value = default_config
        runner = CliRunner()
        result = runner.invoke(check, ["-n", "--no-update"])
        mock_read_state.assert_not_called()
        mock_persist_state.assert_not_called()
        mockfetch_available_updates.assert_not_called()
        assert result.exit_code == 1
        assert result.output == "Error: --no-update and --no-cache options are mutually exclusive\n"

    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state")
    @mock.patch(
        "siun.providers.UpdateProviderPacman.fetch_updates",
        return_value=[],
    )
    @mock.patch("siun.cli_utils.get_config")
    def test_check_quiet_option(
        self, mock_get_config, mockfetch_available_updates, mock_read_state, mock_persist_state, default_config
    ):
        """Test check CLI command with no updates and --quiet option."""
        mock_get_config.return_value = default_config
        runner = CliRunner()
        result = runner.invoke(check, ["-n", "-q"])
        mock_read_state.assert_not_called()
        mock_persist_state.assert_not_called()
        mockfetch_available_updates.assert_called_once()
        assert result.exit_code == 0
        assert result.output == ""

    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state")
    @mock.patch(
        "siun.providers.UpdateProviderPacman.fetch_updates",
        return_value=[],
    )
    @mock.patch("siun.cli_utils.get_config")
    def test_check_no_updates(
        self, mock_get_config, mockfetch_available_updates, mock_read_state, mock_persist_state, default_config
    ):
        """Test check CLI command with no updates."""
        mock_read_state.return_value = EMPTY_STATE
        mock_get_config.return_value = default_config
        runner = CliRunner()
        result = runner.invoke(check)
        mock_read_state.assert_called_once()
        mock_persist_state.assert_not_called()
        mockfetch_available_updates.assert_not_called()
        assert result.exit_code == 0
        assert result.output == "No matches.\n"

    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state")
    @mock.patch(
        "siun.providers.UpdateProviderPacman.fetch_updates",
        return_value=[],
    )
    @mock.patch("siun.cli_utils.get_config", side_effect=ConfigError("failed", config_path=Path("/path/to/siun.toml")))
    def test_check_invalid_config(
        self, mock_get_config, mockfetch_available_updates, mock_read_state, mock_persist_state, default_config
    ):
        """Test get_state CLI command with invalid config."""
        mock_get_config.return_value = default_config
        runner = CliRunner()
        result = runner.invoke(check)
        mock_read_state.assert_not_called()
        mock_persist_state.assert_not_called()
        mockfetch_available_updates.assert_not_called()
        assert result.exit_code == 1
        assert result.output == "Error: failed; config path: /path/to/siun.toml\n"

    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state")
    @mock.patch(
        "siun.providers.UpdateProviderPacman.fetch_updates",
        return_value=[PackageUpdate(name="package", provider="pacman")],
    )
    @mock.patch("siun.cli_utils.get_config")
    def test_check_stale_state(
        self,
        mock_get_config,
        mockfetch_available_updates,
        mock_read_state,
        mock_persist_state,
        default_config,
        state_stale,
    ):
        """Test check CLI command with stale state on disk."""
        mock_read_state.return_value = state_stale
        mock_get_config.return_value = default_config
        runner = CliRunner()
        result = runner.invoke(check)
        mock_read_state.assert_called_once()
        mock_persist_state.assert_called_once()
        mockfetch_available_updates.assert_called_once()
        assert result.exit_code == 0
        assert result.output == "Updates available\n"

    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state")
    @mock.patch(
        "siun.cli.get_updates",
        side_effect=SiunGetUpdatesError("[pacman] Permission denied"),
    )
    @mock.patch("siun.cli_utils.get_config")
    def test_check_with_error_onfetch_available_updates(
        self,
        mock_get_config,
        mockfetch_available_updates,
        mock_read_state,
        mock_persist_state,
        default_config,
        state_stale,
    ):
        """Test check CLI command failing to get available updates."""
        mock_read_state.return_value = state_stale
        mock_get_config.return_value = default_config
        runner = CliRunner()
        result = runner.invoke(check, ["-n"])
        mock_read_state.assert_not_called()
        mock_persist_state.assert_not_called()
        mockfetch_available_updates.assert_called_once()
        assert result.exit_code == 1
        assert result.stdout == ""
        assert result.stderr == "Error: [pacman] Permission denied\n"

    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state")
    @mock.patch(
        "siun.providers.UpdateProviderPacman.fetch_updates",
        return_value=[PackageUpdate(name="package", provider="pacman")],
    )
    @mock.patch("siun.cli_utils.get_config")
    def test_check_with_custom_output_format(
        self,
        mock_get_config,
        mockfetch_available_updates,
        mock_read_state,
        mock_persist_state,
        v2_config_w_custom_format,
        state_stale,
    ):
        """Test check CLI command with custom output format."""
        mock_read_state.return_value = state_stale
        mock_get_config.return_value = v2_config_w_custom_format
        runner = CliRunner()
        result = runner.invoke(check, ["-o", "custom"])
        mock_read_state.assert_called_once()
        mock_persist_state.assert_called_once()
        mockfetch_available_updates.assert_called_once()
        assert result.exit_code == 0
        assert result.output == "Updates available: package\n"

    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state", return_value=False)
    @mock.patch(
        "siun.providers.UpdateProviderPacman.fetch_updates",
        return_value=[PackageUpdate(name="package", provider="pacman")],
    )
    @mock.patch("siun.config._read_config", return_value=tomllib.loads(CONFIG_CUSTOM_STATE_FILE_PATH))
    def test_custom_state_file_path_config(
        self, mock_read_config, mockfetch_available_updates, mock_read_state, mock_persist_state
    ):
        """Test check CLI command with custom state file path."""
        mock_read_state.return_value = False
        runner = CliRunner()
        with mock.patch(
            "siun.config.get_default_config_dir"
        ):  # NOTE: Required because get_config only tries to read the config file when it exists
            result = runner.invoke(check)

        mock_read_config.assert_called_once()
        mock_read_state.assert_called_once_with(Path("/tmp/siun-test-state/state.json"))  # noqa: S108
        mock_persist_state.assert_called_once_with(Path("/tmp/siun-test-state/state.json"))  # noqa: S108
        mockfetch_available_updates.assert_called_once()
        assert result.exit_code == 0
        assert result.output == "Updates available\n"

    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state", return_value=False)
    def test_get_updates_without_cache_or_update(self, mock_read_state, mock_persist_state):
        """
        Test get_updates with no_cache and no_update.

        The CLI should already catch that no_cache and no_update are mutually
        exclusive, but the code shouldn't fail anyway.
        """
        mock_read_state.return_value = False
        result = get_updates(
            no_cache=True,
            no_update=True,
            criteria=[],
            thresholds=[],
            update_providers=[UpdateProvider(name="dummy")],
            cache_min_age_minutes=0,
            state_file_path=Path("/tmp/siun-test-state.json"),  # noqa: S108
        )
        mock_read_state.assert_not_called()
        mock_persist_state.assert_not_called()
        assert result.score == 0

    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch("siun.check.load_state")
    def test_get_updates_with_no_update_and_existing_cache(self, mock_read_state, mock_persist_state):
        """Test get_updates with no_update."""
        config_criteria = [
            CriterionAvailable(name="available", weight=1),
            CriterionCount(name="count", weight=2, count=1),
        ]
        existing_state = Updates(
            last_update="1970-01-01T01:00:00Z",
            state="OK",
            thresholds=[],
            matched_criteria={},
            available_updates=[{"name": "siun", "provider": "pacman"}],
            criteria_settings=[],
        )
        mock_read_state.return_value = existing_state
        result = get_updates(
            no_cache=False,
            no_update=True,
            criteria=config_criteria,
            thresholds=[],
            update_providers=[UpdateProvider(name="dummy")],
            cache_min_age_minutes=0,
            state_file_path=Path("/tmp/siun-test-state.json"),  # noqa: S108
        )
        mock_read_state.assert_called_once()
        mock_persist_state.assert_not_called()
        assert result.available_updates == [PackageUpdate(name="siun", provider="pacman")]

    @mock.patch("siun.cli.INSTALLED_FEATURES", [])
    @mock.patch(
        "siun.providers.UpdateProviderPacman.fetch_updates",
        return_value=[PackageUpdate(name="package", provider="pacman")],
    )
    @mock.patch("siun.cli_utils.get_config")
    def test_check_notification_wo_feature(self, mock_get_config, mockfetch_available_updates, config_w_notification):
        """Test check CLI command with missing notification feature."""
        mock_get_config.return_value = config_w_notification
        runner = CliRunner()
        result = runner.invoke(check, ["-n"])
        mockfetch_available_updates.assert_called_once()
        assert result.exit_code == 1
        assert result.stderr.startswith("Error: notifications require the 'notification' feature") is True

    @pytest.mark.feature_notification
    @mock.patch("siun.notification.UpdateNotification.show")
    @mock.patch(
        "siun.providers.UpdateProviderPacman.fetch_updates",
        return_value=[PackageUpdate(name="package", provider="pacman")],
    )
    @mock.patch("siun.cli_utils.get_config")
    def test_check_notification(self, mock_get_config, mockfetch_available_updates, mock_show, config_w_notification):
        """Test check CLI command with notification."""
        mock_get_config.return_value = config_w_notification
        runner = CliRunner()
        result = runner.invoke(check, ["-n"])
        mockfetch_available_updates.assert_called_once()
        mock_show.assert_called_once()
        assert result.exit_code == 0
        assert result.output == "Updates available\n"

    @pytest.mark.feature_notification
    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch(
        "siun.providers.UpdateProviderPacman.fetch_updates",
        return_value=[PackageUpdate(name="siun", provider="pacman")],
    )
    @mock.patch("siun.check.load_state")
    @mock.patch("siun.notification.UpdateNotification.show")
    @mock.patch("siun.cli_utils.get_config")
    def test_check_notification_below_threshold(
        self,
        mock_get_config,
        mock_show,
        mock_read_state,
        mock_get_updates,
        mock_persist_state,
        config_w_notification_threshold,
        state_stale,
    ):
        """Test state below threshold does not show notification."""
        mock_get_config.return_value = config_w_notification_threshold
        # `available` state is below notification threshold
        mock_read_state.return_value = state_stale
        runner = CliRunner()
        result = runner.invoke(check)
        mock_show.assert_not_called()
        assert result.exit_code == 0
        assert result.output == "Updates available\n"

    @pytest.mark.feature_notification
    @mock.patch("siun.cli.Updates.persist_state")
    @mock.patch(
        "siun.providers.UpdateProviderPacman.fetch_updates",
        return_value=[
            PackageUpdate(name="package", provider="pacman"),
            PackageUpdate(name="other_package", provider="pacman"),
        ],
    )
    @mock.patch("siun.check.load_state")
    @mock.patch("siun.notification.UpdateNotification.show")
    @mock.patch("siun.cli_utils.get_config")
    def test_check_notification_above_threshold(
        self,
        mock_get_config,
        mock_show,
        mock_read_state,
        mock_get_updates,
        mock_persist_state,
        config_w_notification_threshold,
        state_stale,
    ):
        """Test state gte threshold shows notification."""
        mock_get_config.return_value = config_w_notification_threshold
        # `warning` state exceeds notification threshold
        mock_read_state.return_value = state_stale
        runner = CliRunner()
        result = runner.invoke(check)
        assert result.output == "Updates recommended\n"
        mock_show.assert_called_once()
        assert result.exit_code == 0
        assert result.output == "Updates recommended\n"



"""AUR package update provider."""

from __future__ import annotations

import subprocess

from pydantic import ConfigDict

from siun.errors import UpdateProviderError
from siun.models import PackageUpdate
from siun.providers.base import UpdateProvider
from siun.providers.pacman import PACMAN_PATTERN


class UpdateProviderAur(UpdateProvider):
    """Update provider for AUR helpers."""

    name: str = "aur"
    _default_cmds: list[list[str]] = [["aur-check-updates"], ["paru", "--aur", "-Qu"]]
    pattern: str = PACMAN_PATTERN
    exit_code_no_updates: int = 1

    def fetch_updates(self) -> list[PackageUpdate]:
        """Get list of updates from AUR helper."""
        cmd = self.pick_cmd(self._default_cmds)
        try:
            available_updates_run = subprocess.run(  # noqa: S603
                cmd,
                check=True,
                capture_output=True,
                text=True,
                shell=False,
            )
            return self.parse_updates(available_updates_run.stdout.splitlines(), self.pattern)

        except subprocess.CalledProcessError as error:
            # pacman and some AUR helpers return exit code 1 when there are no updates
            if error.returncode != self.exit_code_no_updates:
                raise
        except Exception as error:
            message = f"unexpected error: {error}"
            raise UpdateProviderError(message, self.name) from error

        return []

    model_config = ConfigDict(extra="forbid")



"""Output formatters."""

import json
from enum import Enum
from string import Template
from typing import Never

from click import style as click_style

from siun.models import FormatObject


class OutputFormat(Enum):
    """Output format options."""

    PLAIN = "plain"
    FANCY = "fancy"
    JSON = "json"
    CUSTOM = "custom"


class Formatter:
    """Build output format."""

    @staticmethod
    def format_plain(format_object: FormatObject) -> tuple[str, dict[Never, Never]]:
        """Build simple text format."""
        return format_object.status_text, {}

    @staticmethod
    def format_fancy(format_object: FormatObject) -> tuple[str, dict[str, str]]:
        """Build coloured text format."""
        return format_object.status_text, {"fg": format_object.state_color}

    @staticmethod
    def format_json(format_object: FormatObject) -> tuple[str, dict[Never, Never]]:
        """Build JSON output format."""
        state_dict = {
            "count": format_object.update_count,
            "text_value": format_object.status_text,
            "score": format_object.score,
        }
        return json.dumps(state_dict), {}

    @staticmethod
    def format_custom(format_object: FormatObject, template_string: str) -> tuple[str, dict[Never, Never]]:
        """Build customised output format."""
        format_template = Template(template_string)
        output = format_template.safe_substitute(**format_object.model_dump())
        return output, {}


def get_formatted_state_text(format_object: FormatObject, output_format: OutputFormat, custom_format: str) -> str:
    """Generate formatted output text from update state."""
    formatter = Formatter()
    formatter_kwargs = {}
    if output_format == OutputFormat.CUSTOM:
        formatter_kwargs["template_string"] = custom_format
    formatted_output, format_options = getattr(formatter, f"format_{output_format.value}")(
        format_object, **formatter_kwargs
    )
    return click_style(formatted_output, **format_options)



from .aur import UpdateProviderAur as UpdateProviderAur
from .base import UPDATE_PROVIDER_REGISTRY as UPDATE_PROVIDER_REGISTRY
from .base import UpdateProvider as UpdateProvider
from .flatpak import UpdateProviderFlatpak as UpdateProviderFlatpak
from .generic import UpdateProviderGeneric as UpdateProviderGeneric
from .pacman import UpdateProviderPacman as UpdateProviderPacman



"""Test notification module."""

from unittest import mock

import pytest
from pydantic import ValidationError

from siun.models import FormatObject
from siun.notification import NotificationUrgency, UpdateNotification


class TestNotification:
    """Test UpdateNotification."""

    @mock.patch("siun.notification.UpdateNotification.show")
    def test_defaults_ok(self, mocked_show):
        """Test defaults are enough."""
        notification = UpdateNotification(threshold="available")
        notification.show()

        assert notification.app_name == "siun"
        assert notification.title == "$status_text"
        mocked_show.assert_called_once()

    @pytest.mark.parametrize(
        "urgency, expected_value",
        [
            (None, None),
            ("low", NotificationUrgency.low),
            ("normal", NotificationUrgency.normal),
            ("critical", NotificationUrgency.critical),
        ],
    )
    def test_urgency_validation_ok(self, urgency, expected_value):
        """Test valid urgency values."""
        notification = UpdateNotification(urgency=urgency, threshold="critical")

        assert notification.app_name == "siun"
        assert notification.urgency == expected_value

    @pytest.mark.parametrize(
        "urgency",
        ["invalid_string", "", 31415],
    )
    def test_urgency_validation_fail(self, urgency):
        """Test invalid urgency values."""
        with pytest.raises(ValidationError) as error:
            UpdateNotification(urgency=urgency)

        assert error.value.errors()[0]["msg"].startswith("Value error, input should be a valid urgency")

    def test_template_values(self):
        """Test filling of template variables."""
        template_string_all = (
            "$available_updates | $last_update | $matched_criteria | "
            "$matched_criteria_short | $score | $status_text | $update_count"
        )
        format_obj = FormatObject(
            available_updates="siun",
            last_update="2025-04-09T00:00:00+00:00",
            matched_criteria="available",
            matched_criteria_short="av",
            score=2,
            status_text="Updates available",
            update_count=1,
            state_color="green",
            state_name="AVAILABLE_UPDATES",
        )
        notification = UpdateNotification(
            title="$status_text: $available_updates", message=template_string_all, threshold="available"
        )

        notification.fill_templates(format_obj)

        assert notification.title == "Updates available: siun"
        assert notification.message == "siun | 2025-04-09T00:00:00+00:00 | available | av | 2 | Updates available | 1"



"""Base package update provider and update provider registry."""

from __future__ import annotations

import re
from shutil import which

from pydantic import BaseModel, ConfigDict

from siun.errors import UpdateProviderError
from siun.models import PackageUpdate

UPDATE_PROVIDER_REGISTRY: dict[str, type[UpdateProvider]] = {}


class UpdateProvider(BaseModel):
    """Base provider for available updates."""

    name: str

    def fetch_updates(self) -> list[PackageUpdate]:
        """Scaffolding for fetching list of available updates."""
        raise NotImplementedError

    def parse_updates(self, lines: list[str], pattern: str) -> list[PackageUpdate]:
        """Parse list of available update strings into list of PackageUpdate objects."""
        available_updates: list[PackageUpdate] = []
        for line in lines:
            match = re.match(pattern, line)
            if not match or "name" not in match.groupdict():
                message = f"failed to parse output: {line}"
                raise UpdateProviderError(message, self.name)

            match_dict = match.groupdict()
            available_updates.append(
                PackageUpdate(
                    name=match_dict["name"],
                    old_version=match_dict.get("old_version"),
                    new_version=match_dict.get("new_version"),
                    provider=self.name,
                )
            )

        return available_updates

    def pick_cmd(self, cmds: list[list[str]]) -> list[str]:
        """Pick the first available command from the list."""
        for cmd in cmds:
            if which(cmd[0]) is not None:
                return cmd

        message = f"no suitable command found among: {cmds}"
        raise UpdateProviderError(message, self.name)

    def __init_subclass__(cls, **kwargs: object) -> None:
        """Register subclass in criterion registry."""
        super.__init_subclass__(**kwargs)
        UPDATE_PROVIDER_REGISTRY[cls.name] = cls

    model_config = ConfigDict(extra="allow")



"""CLI commands for siun."""

from collections import defaultdict
from pathlib import Path

import click

from siun import __version__
from siun.check import get_updates
from siun.cli_utils import common_options, load_config_or_exit
from siun.config import SiunConfig
from siun.errors import (
    SiunCLIError,
    SiunGetUpdatesError,
    SiunNotificationError,
)
from siun.formatting import OutputFormat, get_formatted_state_text
from siun.models import NewsEntry
from siun.models.updates import Updates
from siun.news import INSTALLED_FEATURES as INSTALLED_NEWS_FEATURES
from siun.news import (
    format_news_entries,
    load_news_state,
    parse_feed_entries,
    save_news_state,
)
from siun.notification import INSTALLED_FEATURES as INSTALLED_NOTIFICATION_FEATURES

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}
INSTALLED_FEATURES: set[str] = INSTALLED_NOTIFICATION_FEATURES | INSTALLED_NEWS_FEATURES


def _handle_notification(config: SiunConfig, siun_state: Updates) -> None:
    notification = config.notification
    if not notification:
        return None

    if "notification" not in INSTALLED_FEATURES:
        message = "notifications require the 'notification' feature, install with 'pip install siun[notification]'"
        raise SiunNotificationError(message)

    threshold_score = config.mapped_thresholds[notification.threshold].score
    if (
        not siun_state.match
        or (siun_state.last_match and siun_state.match.score <= siun_state.last_match.score)
        or siun_state.match.score < threshold_score
    ):
        return None

    notification.fill_templates(siun_state.format_object)
    if notification.urgency is not None:
        notification.hints = {"urgency": notification.urgency.value}
    notification.show()


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__)
def cli() -> None:  # noqa: D103 # pragma: no cover
    pass


@cli.command()
@click.option("--nocolor", is_flag=True, show_default=True, default=False, help="Do not colorize output.")
@common_options
def news(*, config_path: Path, nocolor: bool) -> None:
    """Show latest news headlines."""
    if "news" not in INSTALLED_FEATURES:
        message = "news require the 'news' feature, install with 'pip install siun[news]'"
        raise SiunCLIError(message)

    config = load_config_or_exit(config_path)
    news_sources = config.news

    last_update_path = config.state_dir / Path("last_news_update")
    load_news_state(news_sources, last_update_path)

    news_entries: dict[str, list[NewsEntry]] = defaultdict(list)
    for source in news_sources:
        feed_title, entries = parse_feed_entries(source)
        news_entries[feed_title].extend(entries)

    save_news_state(news_sources, last_update_path)

    click.echo(format_news_entries(news_entries, nocolor))


@cli.command()
@click.option("--quiet", "-q", is_flag=True, show_default=True, default=False, help="Suppress output.")
@click.option(
    "--no-update", "-U", is_flag=True, show_default=True, default=False, help="Don't get updates, only check."
)
@click.option("--cache/--no-cache", " /-n", show_default=True, default=True, help="Ignore existing state on disk.")
@click.option(
    "--output-format",
    "-o",
    default=OutputFormat.PLAIN,
    type=click.Choice(OutputFormat, case_sensitive=False),
    help="Pick output format for update check.",
)
@common_options
def check(*, config_path: Path, output_format: OutputFormat, cache: bool, no_update: bool, quiet: bool) -> None:
    """Check for urgency of available updates."""
    if no_update and not cache:
        raise SiunCLIError(message="--no-update and --no-cache options are mutually exclusive")

    config = load_config_or_exit(config_path)
    try:
        siun_state = get_updates(
            no_cache=not cache,
            no_update=no_update,
            criteria=config.v2_criteria,
            thresholds=config.sorted_thresholds,
            cache_min_age_minutes=config.cache_min_age_minutes,
            state_file_path=config.state_dir / Path("state.json"),
            update_providers=config.update_providers,
        )
    except SiunGetUpdatesError as error:
        raise SiunCLIError(error.message) from error

    if not quiet:
        formatted_output = get_formatted_state_text(siun_state.format_object, output_format, config.custom_format)
        click.echo(formatted_output)

    try:
        _handle_notification(config, siun_state)
    except SiunNotificationError as error:
        raise SiunCLIError(error.message) from error



"""Internal state of siun and available updates."""

import importlib.util
from importlib.machinery import SourceFileLoader
from pathlib import Path
from typing import Any

from siun.criteria import CriterionAvailable, CriterionCount, CriterionPattern, SiunCriterion
from siun.errors import CriterionError
from siun.models import PackageUpdate, V2Criterion
from siun.models.updates import Updates
from siun.providers import UpdateProvider
from siun.util import get_default_criteria_dir, is_path_world_writable

BUILTIN_CRITERIA: dict[str, SiunCriterion] = {
    "available": CriterionAvailable(),
    "count": CriterionCount(),
    "pattern": CriterionPattern(),
}
EXPECTED_CLASS = "SiunCriterion"

Updates.model_rebuild()


def _load_user_criteria(*, criteria_settings: list[V2Criterion], include_path: Path | None = None) -> dict[str, Any]:
    """Load user criteria."""
    user_criteria: dict[str, SiunCriterion] = {}
    if not include_path:
        include_path = get_default_criteria_dir()

    if not include_path.exists() or not include_path.is_dir():
        return user_criteria

    if is_path_world_writable(include_path):
        message = (
            f"Criteria directory '{include_path}' is world-writable. "
            "Please change its permissions to be more restrictive."
        )
        raise ImportError(message)

    # Get list of enabled user criteria from config
    enabled_criteria = {criterion.name for criterion in criteria_settings if criterion.weight != 0}

    for f_name in include_path.iterdir():
        # Only load enabled user criteria
        if f_name.suffix != ".py" or f_name.stem not in enabled_criteria:
            continue  # Skip non-Python files
        file_path = include_path / f_name
        py_mod_loader = SourceFileLoader(file_path.stem, file_path.as_posix())
        py_mod_spec = importlib.util.spec_from_loader(py_mod_loader.name, py_mod_loader)
        if py_mod_spec is None:
            message = "Could not create module specification for source file loader"
            raise ImportError(message)
        py_mod = importlib.util.module_from_spec(py_mod_spec)
        py_mod_loader.exec_module(py_mod)
        if hasattr(py_mod, EXPECTED_CLASS):
            # Check inheritance
            class_obj = getattr(py_mod, EXPECTED_CLASS)
            if not hasattr(class_obj, "is_fulfilled") or not callable(class_obj.is_fulfilled):
                continue  # An error is raised later if any configured criteria could not be loaded
            class_inst = py_mod.SiunCriterion()
            user_criteria[file_path.stem] = class_inst

    return user_criteria


def load_state(state_file_path: Path) -> Updates | None:
    """Read state from disk."""
    if not state_file_path.exists():
        return None

    with Path.open(state_file_path) as update_file:
        return Updates.model_validate_json(update_file.read())


def get_package_updates(update_providers: list[UpdateProvider]) -> list[PackageUpdate]:
    """Fetch available package updates."""
    package_updates: list[PackageUpdate] = []
    for provider in update_providers:
        package_updates.extend(provider.fetch_updates())

    return package_updates


def get_merged_criteria(criteria_settings: list[V2Criterion]) -> dict[str, SiunCriterion]:
    """Return merged built-in and user criteria."""
    try:
        criteria = BUILTIN_CRITERIA.copy()
        user_criteria = _load_user_criteria(criteria_settings=criteria_settings)
        criteria.update(user_criteria)
    except Exception as error:
        message = f"Error loading criteria: {error}"
        raise CriterionError(message, None) from error

    return criteria



"""Models dealing with package updates."""

from __future__ import annotations

import datetime
from pathlib import Path
from typing import Any

from pydantic import BaseModel, ConfigDict

from siun.criteria import SiunCriterion
from siun.models.criteria import V2Criterion
from siun.models.formatting import ClickColor, FormatObject
from siun.models.thresholds import V2Threshold


class PackageUpdate(BaseModel):
    """Struct representing an available update."""

    name: str
    old_version: str | None = None
    new_version: str | None = None
    provider: str

    model_config = ConfigDict(extra="allow")


class Updates(BaseModel):
    """Internal state struct."""

    criteria_settings: list[V2Criterion] = []
    thresholds: list[V2Threshold] = []
    available_updates: list[PackageUpdate] = []
    matched_criteria: dict[str, dict[str, Any]] = {}
    last_update: datetime.datetime = datetime.datetime.now(tz=datetime.UTC)
    match: V2Threshold | None = None
    last_match: V2Threshold | None = None

    def touch(self) -> None:
        """Set last_update to now."""
        self.last_update = datetime.datetime.now(tz=datetime.UTC)

    @property
    def score(self) -> int:
        """Calculate score from criteria weights."""
        return sum([criterium["weight"] for criterium in self.matched_criteria.values()])

    @property
    def count(self) -> int:
        """Get count of available updates."""
        return len(self.available_updates)

    @property
    def color(self) -> ClickColor:
        """Get color of matched threshold."""
        if not self.match:
            return ClickColor.reset
        return self.match.color

    @property
    def text_value(self) -> str:
        """Get text value of matched threshold."""
        if not self.match:
            return "No matches."
        return self.match.text

    @property
    def format_object(self) -> FormatObject:
        """Provide prepared values for formatters."""
        return FormatObject(
            available_updates=", ".join([update.name for update in self.available_updates]),
            last_update=self.last_update.replace(microsecond=0).isoformat(),
            matched_criteria=", ".join(self.matched_criteria.keys()),
            matched_criteria_short=",".join([match["name_short"] for match in self.matched_criteria.values()]),
            score=self.score,
            status_text=self.text_value,
            update_count=self.count,
            state_color=self.color.value,
            state_name=self.text_value,
        )

    def evaluate(
        self,
        criteria: dict[str, SiunCriterion],
        available_updates: list[PackageUpdate] | None = None,
    ) -> None:
        """Update state of updates. Criteria must be passed in."""
        self.touch()
        if available_updates is None:
            available_updates = []

        self.available_updates = available_updates
        self.matched_criteria = {}  # Reset matches

        # Check criteria
        for crit in self.criteria_settings:
            if crit and crit.weight == 0:
                continue  # Skip criteria with weight 0
            try:
                user_criteria_settings = crit.model_dump(exclude={"name", "short_name"})
                if crit.name not in criteria:
                    message = (
                        f"Configured criterion '{crit.name}' was not loaded. "
                        "Likely reasons:\n"
                        "- Missing or misplaced criterion file\n"
                        "- Criterion class missing or misnamed\n"
                        "- 'is_fulfilled' method not implemented\n"
                        "Check your criteria directory and configuration."
                    )
                    from siun.errors import CriterionError

                    raise CriterionError(message, crit.name)
                if criteria[crit.name].is_fulfilled(
                    user_criteria_settings, [update.name for update in available_updates]
                ):
                    self.matched_criteria[crit.name] = user_criteria_settings
            except Exception as error:
                from siun.errors import CriterionError

                crit_settings = crit.model_dump()
                import traceback

                tb = traceback.format_exc()
                message = f"Criterion settings: {crit_settings}\nTraceback:\n{tb}"
                raise CriterionError(message, crit.name) from error

        for threshold in self.thresholds:
            if self.score >= threshold.score:
                self.match = threshold
                break
        else:
            # Reset match if no thresholds matched
            self.match = None

    def persist_state(self, state_file_path: Path) -> None:
        """Write state to disk."""
        from siun.util import safely_write_to_disk

        return safely_write_to_disk(
            content=self.model_dump_json(exclude={"thresholds", "last_match", "matched_criteria", "criteria_settings"}),
            target_path=state_file_path,
        )



"""Test formatting module."""

import pytest

from siun.formatting import Formatter


@pytest.mark.parametrize(
    "format_object_fixture,expected_output",
    [
        ("format_object_ok", "Ok"),
        ("format_object_available", "Updates available"),
        ("format_object_recommended", "Updates recommended"),
        ("format_object_required", "Updates required"),
    ],
)
def test_plain(request, format_object_fixture, expected_output):
    """Test plain formatter."""
    format_object = request.getfixturevalue(format_object_fixture)
    output, output_kwargs = Formatter.format_plain(format_object)
    assert output == expected_output
    assert output_kwargs == {}


@pytest.mark.parametrize(
    "format_object_fixture,expected_output,expected_kwargs",
    [
        ("format_object_ok", "Ok", {"fg": "green"}),
        ("format_object_available", "Updates available", {"fg": "blue"}),
        ("format_object_recommended", "Updates recommended", {"fg": "yellow"}),
        ("format_object_required", "Updates required", {"fg": "red"}),
    ],
)
def test_fancy(request, format_object_fixture, expected_output, expected_kwargs):
    """Test fancy formatter."""
    format_object = request.getfixturevalue(format_object_fixture)
    output, output_kwargs = Formatter.format_fancy(format_object)
    assert output == expected_output
    assert output_kwargs == expected_kwargs


@pytest.mark.parametrize(
    "format_object_fixture,expected_output,expected_kwargs",
    [
        ("format_object_ok", '{"count": 0, "text_value": "Ok", "score": 0}', {}),
        ("format_object_available", '{"count": 0, "text_value": "Updates available", "score": 0}', {}),
        ("format_object_recommended", '{"count": 0, "text_value": "Updates recommended", "score": 0}', {}),
        ("format_object_required", '{"count": 0, "text_value": "Updates required", "score": 0}', {}),
    ],
)
def test_json(request, format_object_fixture, expected_output, expected_kwargs):
    """Test JSON formatter."""
    format_object = request.getfixturevalue(format_object_fixture)
    output, output_kwargs = Formatter.format_json(format_object)
    assert output == expected_output
    assert output_kwargs == expected_kwargs


def test_json_handles_count_correctly(format_object_factory):
    """Test count for JSON formatter."""
    format_object = format_object_factory(
        available_updates="pkg1, pkg2",
        status_text="Updates recommended",
        update_count=42,
        state_color="yellow",
        state_name="Updates recommended",
    )
    output, output_kwargs = Formatter.format_json(format_object)
    assert output == '{"count": 42, "text_value": "Updates recommended", "score": 0}'
    assert output_kwargs == {}


def test_json_handles_score_correctly(format_object_factory):
    """Test score for JSON formatter."""
    format_object = format_object_factory(
        available_updates="pkg1, important-pkg",
        matched_criteria="available, pattern",
        matched_criteria_short="av,pa",
        score=3,
        status_text="Updates recommended",
        update_count=2,
        state_color="yellow",
        state_name="Updates recommended",
    )
    output, output_kwargs = Formatter.format_json(format_object)
    assert output == '{"count": 2, "text_value": "Updates recommended", "score": 3}'
    assert output_kwargs == {}


@pytest.mark.parametrize(
    "format_object_fixture,template_string,expected_output",
    [
        ("format_object_ok", "$status_text: $available_updates", "Ok: "),
        ("format_object_available", "$status_text: $available_updates", "Updates available: "),
        ("format_object_recommended", "$status_text: $available_updates", "Updates recommended: "),
        ("format_object_required", "$status_text: $available_updates", "Updates required: "),
        (
            None,
            "$available_updates | $last_update | $matched_criteria | $matched_criteria_short | $score | $status_text | $update_count",
            "dummy | 2025-01-09T00:00:00+00:00 | available, count | av,co | 7 | Updates available | 1",
        ),
    ],
)
def test_custom(request, format_object_fixture, template_string, expected_output, format_object_factory):
    """Test custom formatter."""
    if format_object_fixture is not None:
        obj = request.getfixturevalue(format_object_fixture)
    else:
        obj = format_object_factory(
            available_updates="dummy",
            last_update="2025-01-09T00:00:00+00:00",
            matched_criteria="available, count",
            matched_criteria_short="av,co",
            score=7,
            status_text="Updates available",
            update_count=1,
            state_color="blue",
            state_name="Updates available",
        )
    output, output_kwargs = Formatter.format_custom(obj, template_string)
    assert output == expected_output
    assert output_kwargs == {}



"""Common utils."""

import shutil
import stat
import tempfile
from os import environ
from pathlib import Path


def get_default_state_dir() -> Path:
    """
    Provide default value for state_dir setting.

    By default, siun will try the following in order:
    1. `$XDG_STATE_HOME/siun/state.json`
    2. `$HOME/.local/state/siun/state.json`
    """
    return Path(environ.get("XDG_STATE_HOME", Path.home() / Path(".local/state"))) / Path("siun")


def get_default_config_dir() -> Path:
    """
    Provide default value for criteria_path setting.

    By default, siun will try the following in order:
    1. `$XDG_CONFIG_HOME/siun`
    2. `$HOME/.config/siun`
    """
    return Path(environ.get("XDG_CONFIG_HOME", Path.home() / Path(".config"))) / Path("siun")


def get_default_criteria_dir() -> Path:
    """
    Provide default value for criteria_path setting.

    By default, siun will try the following in order:
    1. `$XDG_CONFIG_HOME/siun/criteria`
    2. `$HOME/.config/siun/criteria`
    """
    return get_default_config_dir() / Path("criteria")


def safely_write_to_disk(*, content: str, target_path: Path) -> None:
    """
    Safely write to disk.

    Avoids partially written state file (and therefore invalid JSON) by
    creating a temporary file first and only replacing the state file once
    the writing operation is done.
    """
    if not Path.exists(target_path.parent):
        # Create parent directory for state file path if it doesn't exist
        Path.mkdir(Path(target_path.parent), parents=True, exist_ok=True)

    with tempfile.NamedTemporaryFile(mode="w+") as update_file:
        update_file.write(content)
        update_file.flush()
        shutil.copy(update_file.name, target_path)


def is_path_world_writable(path: Path) -> bool:
    """Check if a given path is world-writable."""
    return bool(path.stat().st_mode & stat.S_IWOTH)



"""Test news command."""

import tempfile
from pathlib import Path
from unittest import mock

import pytest
from click.testing import CliRunner

from siun.cli import news
from siun.models import NewsProvider

DUMMY_FEED_DATA = """<?xml version="1.0" encoding="UTF-8"?>
<rss version="2.0">
  <channel>
    <title>PyPI recent updates for siun</title>
    <link>https://pypi.org/project/siun/</link>
    <description>Recent updates to the Python Package Index for siun</description>
    <language>en</language>    <item>
      <title>1.5.1</title>
      <link>https://pypi.org/project/siun/1.5.1/</link>
      <description>Report urgency of package upgrades</description>
<author>t4k1t+dev@protonmail.com</author>      <pubDate>Sat, 20 Sep 2025 18:09:49 GMT</pubDate>
    </item>    <item>
      <title>1.5.0</title>
      <link>https://pypi.org/project/siun/1.5.0/</link>
      <description>Report urgency of package upgrades</description>
<author>t4k1t+dev@protonmail.com</author>      <pubDate>Fri, 30 May 2025 12:58:56 GMT</pubDate>
    </item>    <item>
      <title>1.4.1</title>
      <link>https://pypi.org/project/siun/1.4.1/</link>
      <description>Report urgency of package upgrades</description>
<author>t4k1t+dev@protonmail.com</author>      <pubDate>Mon, 12 May 2025 18:06:28 GMT</pubDate>
    </item> </channel>
</rss>"""


class TestNewsCommand:
    """Test news command."""

    @pytest.mark.feature_news
    @pytest.mark.usefixtures("os_path_isfile_patch")
    @mock.patch("siun.cli_utils.get_config")
    @mock.patch("siun.cli.save_news_state")
    def test_news_command(self, mock_save_news_state, mock_get_config):
        """Test news CLI command with one source."""
        import feedparser

        dummy_news_source = [mock.Mock(url="dummy-url", title=None, max_items=3)]
        dummy_config = mock.Mock()
        dummy_config.news = dummy_news_source
        with tempfile.TemporaryDirectory() as tmpdir:
            dummy_config.state_dir = Path(tmpdir)
            mock_get_config.return_value = dummy_config

            runner = CliRunner()
            # Let the library parse the dummy data so it can be substituted in the test call below
            parsed_feed = feedparser.parse(DUMMY_FEED_DATA)
            with mock.patch("siun.news.feedparser.parse", return_value=parsed_feed) as mock_feedparser_parse:
                result = runner.invoke(news, [])

                mock_feedparser_parse.assert_called_once()
                assert mock_feedparser_parse.call_args[0][0] == "dummy-url"
                assert result.exit_code == 0
                assert "PyPI recent updates for siun" in result.output
                # Titles
                assert "- 1.5.1" in result.output
                assert "- 1.5.0" in result.output
                assert "- 1.4.1" in result.output
                # Links
                assert "https://pypi.org/project/siun/1.5.1" in result.output
                assert "https://pypi.org/project/siun/1.5.0" in result.output
                assert "https://pypi.org/project/siun/1.4.1" in result.output
                # Publish dates
                assert "2025-09-20" in result.output
                assert "2025-05-30" in result.output
                assert "2025-05-12" in result.output

    @pytest.mark.feature_news
    @mock.patch("siun.cli_utils.get_config")
    def test_news_feature_not_installed(self, mock_get_config):
        """Test error when news feature is not installed."""
        dummy_config = mock.Mock()
        dummy_config.news = [mock.Mock(url="dummy-url", title=None, max_items=3)]
        with tempfile.TemporaryDirectory() as tmpdir, mock.patch("siun.cli.INSTALLED_FEATURES", set()):
            dummy_config.state_dir = Path(tmpdir)
            mock_get_config.return_value = dummy_config

            runner = CliRunner()
            result = runner.invoke(news, [])
            assert result.exit_code != 0
            assert "news require the 'news' feature" in result.output

    @pytest.mark.feature_news
    @mock.patch("siun.cli_utils.get_config")
    def test_news_command_no_sources(self, mock_get_config):
        """Test news CLI command with no sources configured."""
        dummy_config = mock.Mock()
        dummy_config.news = []
        with tempfile.TemporaryDirectory() as tmpdir:
            dummy_config.state_dir = Path(tmpdir)
            mock_get_config.return_value = dummy_config

            runner = CliRunner()
            result = runner.invoke(news, [])

        assert result.exit_code == 0
        assert "No new entries" in result.output or result.output.strip() == ""

    @pytest.mark.feature_news
    @mock.patch("siun.cli_utils.get_config")
    def test_news_command_empty_feed(self, mock_get_config):
        """Test news CLI command with a feed that returns no entries."""
        import feedparser

        dummy_news_source = [
            NewsProvider(
                url="/tmp/siun-tests/dummy-feed.xml",  # noqa: S108
                max_items=3,
                etag=None,
                last_modified=None,
            )
        ]
        dummy_config = mock.Mock()
        dummy_config.news = dummy_news_source
        dummy_config.state_dir = Path("/tmp/siun-tests")  # noqa: S108
        mock_get_config.return_value = dummy_config

        runner = CliRunner()
        parsed_feed = feedparser.parse(DUMMY_FEED_DATA)
        parsed_feed.feed = {"title": "Empty Feed"}
        parsed_feed.entries = []
        with mock.patch("siun.news.feedparser.parse", return_value=parsed_feed):
            result = runner.invoke(news, [])
            assert result.exit_code == 0
            assert "Empty Feed" in result.output
            assert "No new entries" in result.output

    @pytest.mark.feature_news
    @mock.patch("siun.cli_utils.get_config")
    @mock.patch("siun.cli.save_news_state")
    @mock.patch("siun.cli.load_news_state")
    def test_news_command_nocolor(self, mock_load_news_state, mock_save_news_state, mock_get_config):
        """Test news CLI command with --nocolor option."""
        import feedparser

        dummy_news_source = [mock.Mock(url="dummy-url", title=None, max_items=1)]
        dummy_config = mock.Mock()
        dummy_config.news = dummy_news_source
        with tempfile.TemporaryDirectory() as tmpdir:
            dummy_config.state_dir = Path(tmpdir)
            mock_get_config.return_value = dummy_config

            runner = CliRunner()
            parsed_feed = feedparser.parse(DUMMY_FEED_DATA)
            with mock.patch("siun.news.feedparser.parse", return_value=parsed_feed):
                result = runner.invoke(news, ["--nocolor"])
                assert result.exit_code == 0
                # Should not contain ANSI color codes
                assert "\x1b[" not in result.output
                assert "PyPI recent updates for siun" in result.output


@pytest.fixture(autouse=True)
def os_path_isfile_patch(monkeypatch):
    """Patch os.path.isfile to always return True."""
    monkeypatch.setattr("os.path.isfile", lambda path: True)
